"""
Contractor which renders a text and a cursor on a page on a screen.
All dimensions on the screen are measured in points (1/72 inch a point).
"""

# author R.N.Bosworth

# version 14 Mar 23  15:13

from guibits1_0 import coloring, cursoring, font_styling
from . import html_parsing
from . import looking_ahead
import math
from . import page_laying_out
from guibits1_0 import painting
from . import string_parsing
from . import text_tokenizing
from . import texting
import time
from guibits1_0 import type_checking2_0
from guibits1_0 import unicoding3_0, windowing, writing

"""
Copyright (C) 2018,2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.

"""
""" 
  Token Grammar version 11 Feb 2019   14:43

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::==  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive.
    

  Text  ::==  Token* EndTag
  
  Token  ::==  Word | Separator | NewLine | InternalTag
  
  EndTag  ::==  < e n d >
  
  Word  ::==  NonSpaceChar NonSpaceChar*
  
  NonSpaceChar is any HtmlCharacter except LF, SPACE, TAB, FF or CR
  
  Separator  ::==  SPACE (U+0020)
  
  Newline  ::==  LF (U+000A)
  
  InternalTag  ::==   (this will be extended later)
  
  Note: Rule for HtmlCharacter is given in HTMLGrammar.
"""

# exposed constants
# -----------------

TO_END = 999  # > end of text(?)

# exposed types
# -------------

"""
class TextAndScreenCursorPosition:
  """ """ cursor position in the text and on the page """ """
  pass
"""

class CursorPosition:
  """ cursor position in the text and on the page """
  pass
  

class Page:
  """  page which is to be rendered """
  pass
  
  
# exposed procedures
# ------------------

def collapse_to_end(curpos):
  """
  pre:
    curpos = current CursorPosition for start and end of cursor
    
  post:
    cursor is thin and at end position
  
  test:
    curpos = None
    curpos = fat cursor
    curpos = thin cursor
  """
  type_checking2_0.check_derivative(curpos,CursorPosition)
  curpos._start_text_position.line_offset = curpos._end_text_position.line_offset
  curpos._start_text_position.code_point_offset = \
    curpos._end_text_position.code_point_offset
  curpos._start_page_position.x_offset = curpos._end_page_position.x_offset
  curpos._start_page_position.y_offset = curpos._end_page_position.y_offset


def collapse_to_start(curpos):
  """
  pre:
    curpos = current CursorPosition for start and end of cursor
    
  post:
    cursor is thin and at start position
  
  test:
    curpos = None
    curpos = fat cursor
    curpos = thin cursor
  """
  type_checking2_0.check_derivative(curpos,CursorPosition)
  curpos._end_text_position.line_offset = curpos._start_text_position.line_offset
  curpos._end_text_position.code_point_offset = \
    curpos._start_text_position.code_point_offset
  curpos._end_page_position.x_offset = curpos._start_page_position.x_offset
  curpos._end_page_position.y_offset = curpos._start_page_position.y_offset


def is_thin(curpos):
  """
  pre:
    curpos = current CursorPosition for start and start of cursor
    
  post:
    return True iff cursor is thin
    
  test:
     curpos = None
     curpos = ((123,234),(124,235)
     curpos = ((123,234),(124,234)
     curpos = ((123,234),(123,235)
     curpos = ((123,234),(123,234)
  """
  type_checking2_0.check_derivative(curpos,CursorPosition)
  return (curpos._start_text_position.line_offset == \
          curpos._end_text_position.line_offset and \
          curpos._start_text_position.code_point_offset == \
          curpos._end_text_position.code_point_offset)
  

def move_cursor_down(curpos,pg):
  """ 
  pre:
    curpos = current CursorPosition for start and end of cursor
    pg = Page for this cursor
    
  post:
    if possible,
      cursor has been collapsed to its end position, if necessary
      cursor's page position has been moved downwards by one line,
        preserving the desired x-offset to ensure a linear path, as far as possible
      cursor's text position has been set to the equivalent of the page position
      true has been returned
    else
      false has been returned
  """
  """ 
  pre:
    curpos._end_page_position.y_offset = current y-offset of end of cursor in page
    curpos._desired_x_offset = desired x-offset of cursor in page, to give linear path
    pg._EOT_position.y_offset = y-offset of last line of page
    pg._line_list = list of x-offsets and lines for this page
    pg._back_map = map from each glyph-interstice of the page 
                     to the equivalent text position

  post:
    either:
      curpos._start_page_position.y_offset = new y-offset of cursor in page
      curpos._start_page_position.x_offset = desired x-offset of cursor in page
      curpos._desired_x_offset = desired x-offset of cursor in page, to give linear path
      curpos._update_x_offset = false, so _desired_x_offset is used by render_text
      curpos._start_text_position.line_offset = equivalent line offset of page position
      curpos._start_text_position.code_point_offset = 
        equivalent code point offset of page position
      curpos._end_page_position.y_offset = new y-offset of cursor in page
      curpos._end_page_position.x_offset = desired x-offset of cursor in page
      curpos._end_text_position.line_offset = equivalent line offset of page position
      curpos._end_text_position.code_point_offset = 
        equivalent code point offset of page position
      true has been returned
    or:
      false has been returned
  """
  """ 
  test:
    curpos = None
    curpos valid
      pg = None
      pg valid
        font name = Courier New
          font size = 12
            horizontal indent = 72 (1")
              vertical indent = 72 (1")
                page width = 288 (4") (text width = 144 i.e. 2" i.e. 20 glyphs)
                  pg._line_list =
                    "In Xanadu did Kubla"
                    "Khan"
                    "A stately pleasure"
                    "dome decree"
                  pg._back_map = map back to text:
                    "In Xanadu did Kubla Khan"
                    "A stately pleasure dome decree"
                  pg._page_position.y_offset = 36.0 (fourth line)
                    page height = 180 (2.5") (visible text height = 36 i.e. 0.5" i.e. 3 lines)
                      curpos._page position.y_offset = 24.0 (third line)
                    page height = 192 (2.67") (visible text height = 48 i.e. 0.67" i.e. 4 lines)
                      curpos._page position.y_offset = 36.0 (fourth line)
                      curpos._page position.y_offset = 24.0 (third line)
                        curpos._desired_x_offset = 129.6 (end of third line)
                        curpos._desired_x_offset = 28.8 (glyph interstice 4 on third line)
  """
  type_checking2_0.check_derivative(curpos,CursorPosition)
  type_checking2_0.check_derivative(pg,Page)
  if curpos._end_page_position.y_offset > (pg._text_height - 2.0*pg._font_size):
    #new position off visible text of page
    return False
  elif _equals_float(curpos._end_page_position.y_offset,pg._EOT_position.y_offset):
    # last non-blank line of page
    return False
  else:
    curpos._end_page_position.y_offset += pg._font_size
    if curpos._update_x_offset:
      curpos._desired_x_offset = curpos._end_page_position.x_offset
    else:
      curpos._end_page_position.x_offset = curpos._desired_x_offset
    curpos._update_x_offset = False  # i.e. keep using _desired_x_offset
    (plo,gio,xoff,yoff) = nearest_glyph_interstice_of(curpos._end_page_position.x_offset,curpos._end_page_position.y_offset,pg)
    (tlo,cpo) = pg._back_map[plo][gio]
    # set up text and page position for thin cursor
    set_thin_cursor_text_and_page(curpos,tlo,cpo,xoff,yoff)
    render_thin_cursor(curpos,pg)
    return True
     

def move_cursor_up(curpos,pg):
  """ 
  pre:
    curpos = current CursorPosition for start and end of cursor
    pg = Page for this cursor
    
  post:
    if possible,
      cursor has been collapsed to its start position, if necessary
      cursor's page position has been moved upwards by one line,
        preserving the desired x-offset to ensure a linear path, as far as possible
      cursor's text position has been set to the equivalent of the page position
      true has been returned
    else
      false has been returned
  """
  """  
  pre:
    curpos._start_page_position.y_offset = current y-offset of cursor in page
    curpos._desired_x_offset = desired x-offset of cursor in page, to give linear path
    pg._page_position.y_offset = y-offset of last line of page
    pg._line_list = list of x-offsets and lines for this page
    pg._back_map = map from each glyph-interstice of the page to the equivalent 
                     text position

  post:
    either:
      curpos._start_page_position.y_offset = new y-offset of cursor in page
      curpos._start_page_position.x_offset = desired x-offset of cursor in page
      curpos._desired_x_offset = desired x-offset of cursor in page, to give linear path
      curpos._update_x_offset = false, so _desired_x_offset is used by render_text
      curpos._start_text_position.line_offset = equivalent line offset of page position
      curpos._start_text_position.code_point_offset = equivalent code point offset of page 
                                                 position
      curpos._end_page_position.y_offset = new y-offset of cursor in page
      curpos._end_page_position.x_offset = desired x-offset of cursor in page
      curpos._desired_x_offset = desired x-offset of cursor in page, to give linear path
      curpos._update_x_offset = false, so _desired_x_offset is used by render_text
      curpos._end_text_position.line_offset = equivalent line offset of page position
      curpos._end_text_position.code_point_offset = equivalent code point offset of page 
                                                 position
      true has been returned
    or:
      false has been returned
  """
  """
  test:
    curpos = None
    curpos valid
      curpos = None
      curpos valid
        pg = None
        pg valid
          font name = Courier New
            font size = 12
              horizontal indent = 72 (1")
                vertical indent = 72 (1")
                  page width = 288 (4") (text width = 144 i.e. 2" i.e. 20 glyphs)
                    pg._line_list =
                      "In Xanadu did Kubla"
                      "Khan"
                      "A stately pleasure"
                      "dome decree"
                    pg._back_map = map back to text:
                      "In Xanadu did Kubla Khan"
                      "A stately pleasure dome decree"
                      page height = 192 (2.67") (visible text height = 48 i.e. 0.67" i.e. 4 lines)
                        tascp._page position.y_offset = 11.9 (just above second line)
                        tascp._page position.y_offset = 12.0 (second line)
                          tascp._desired_x_offset = 28.8 (glyph interstice 4 on second line)
                          tascp._desired_x_offset = 36.0 (beyond end of second line)
  """
  type_checking2_0.check_derivative(curpos,CursorPosition)
  type_checking2_0.check_derivative(pg,Page)
  if curpos._start_page_position.y_offset >= pg._font_size:
    curpos._start_page_position.y_offset -= pg._font_size
    if curpos._update_x_offset:
      curpos._desired_x_offset = curpos._start_page_position.x_offset
    else:
      curpos._start_page_position.x_offset = curpos._desired_x_offset
    curpos._update_x_offset = False  # i.e. keep using _desired_x_offset
    (plo,gio,xoff,yoff) = nearest_glyph_interstice_of(curpos._start_page_position.x_offset,curpos._start_page_position.y_offset,pg)
    (tlo,cpo) = pg._back_map[plo][gio]
    # set up text and page position for thin cursor
    set_thin_cursor_text_and_page(curpos,tlo,cpo,xoff,yoff)
    render_thin_cursor(curpos,pg)
    return True
  else:
    return False
    
  
def nearest_glyph_interstice_of(x,y,pg):
  """
  pre:
    (x,y) = desired (x,y) position of the cursor on the Page, in points
              from the tlh corner of the text
    pg = Page on which the nearest glyph interstice is to be found
    
  post:
    the nearest glyph interstice has been returned, as quadruple 
      (page line offset as int, glyph interstice offset in line as int,
       x-offset in text as float, y-offset in text as float)
      
  note: this procedure takes time O(n log n) where n is the number of
          glyphs in the line of the cursor
      
  test:
    pg._font_name = "Courier New"
      pg._font_styles = {}
        pg._font size = 12
          pg._line_list = [(0.0,"")]
            y = 24.0 (line offset 2)
            y = 12.0 (line offset 1)
            y = 0.0 (line offset 0)
            y = -12.0 (line offset -1)
          pg._line_list = [(0.0,""),(72.0,"frodo")]
            y == 6.0 (halfway down first line
              x = 0.0 (at end of text)
              x = 10.0 (well past end of text)
            y == 18.0 (halfway down 2nd line)
              x = 0.0 (well before text)
              x = 71.9 (just before text)
              x = 72.0 (just at start of text)
              x = 72.1 (just after start of text)
              x = 120.0 (well after text)
              x = 108.1 (just after text)
              x = 108.0 (just at end of text)
              x = 107.9 (just before end of text)
              x = 90.0 (at middle of text)
  """
  page_line_offset = math.floor(y/pg._font_size)
  if page_line_offset < 0:
    page_line_offset = 0
  if page_line_offset >= len(pg._line_list):
    page_line_offset = len(pg._line_list) - 1
  (line_start_offset,line) = pg._line_list[page_line_offset]
  x_in_line = x - line_start_offset
  
  # lo and hi are the bounds of the line, relative to the start of the line
  lo = 0
  lo_off = 0.0
  hi = len(line)
  hi_off = writing.width_in_points_of(pg._my_window,line,pg._font_name,pg._font_styles,pg._font_size)
  # limit cursor glyph interstice to line
  if x_in_line <= lo_off:
    page_glyph_offset = lo
    x_offset_in_line = lo_off
  elif x_in_line >= hi_off:
    page_glyph_offset = hi
    x_offset_in_line = hi_off
  else:
    while hi > lo+1:  # must terminate as hi-lo halves each time
      mid = (lo+hi)//2  # floor division
      mid_off = writing.width_in_points_of(pg._my_window,line[:mid],pg._font_name,pg._font_styles,pg._font_size)
      if x_in_line >= mid_off:
        lo = mid
        lo_off = mid_off
      else:
        hi = mid
        hi_off = mid_off
    # hi == lo+1, lo_off <= x_in_line < hi_off
    lo_distance = x_in_line - lo_off
    hi_distance = hi_off - x_in_line
    if lo_distance <= hi_distance:
      page_glyph_offset = lo
      x_offset_in_line = lo_off
    else:
      page_glyph_offset = hi
      x_offset_in_line = hi_off
  return (page_line_offset, \
          page_glyph_offset, \
          line_start_offset+x_offset_in_line, \
          page_line_offset*pg._font_size)
  
  
def new_cursor_position():
  """ 
  pre:
    true
    
  post:
    new CursorPosition with undefined position has been returned

  test:
    once thru
  """
  curpos = CursorPosition()
  curpos._start_text_position = _new_text_position()
  curpos._end_text_position = _new_text_position()
  curpos._start_page_position = _new_page_position()
  curpos._end_page_position = _new_page_position()
  curpos._desired_x_offset =  -1.0
  curpos._update_x_offset = True
  curpos._start_x_offset = -1.0
  curpos._end_x_offset = -1.0
  curpos._in_fat_cursor = False
  curpos._cursor_rectangle_pending = False
  return curpos
  

def new_page(win,w,h,hi,vi,fn,fss,fsize,c):
  """ 
  pre:
    win = windowing.Window with which this page is associated
    w = width of this page in points, as float
    h = height of this page in points, as float
    hi = horizontal indent of this page in points, as float
    vi = vertical indent of this page in points, as float
    fn = font name for this page, as str
    fss = font styles for this page, as font_styling.FontStyleSet
    fsize = font size for this page, as float
    c = font color for this page
    
  post:
    a new page with the specified dimensions and properties has been returned

  test:
    once thru
  """
  """
  post:
    EOT position has been set to (0,0)
    new empty line_list and back_map have been set up
  """
  type_checking2_0.check_derivative(win,windowing.Window)
  type_checking2_0.check_identical(w,float)
  type_checking2_0.check_identical(h,float)
  type_checking2_0.check_identical(hi,float)
  type_checking2_0.check_identical(vi,float)
  type_checking2_0.check_derivative(fn,str)
  type_checking2_0.check_derivative(fss,font_styling.FontStyles)
  type_checking2_0.check_identical(fsize,float)
  type_checking2_0.check_derivative(c,coloring.Color)
  pg = Page()
  pg._my_window = win
  pg._width = w
  pg._height = h
  pg._horizontal_indent = hi
  pg._vertical_indent = vi
  pg._text_width = pg._width - 2.0*pg._horizontal_indent
  pg._text_height = pg._height - 2.0*pg._vertical_indent
  pg._current_line = unicoding3_0.new_string()
  pg._line_width = 0.0
  pg._current_alignment = texting.Alignment.BEGIN
  pg._font_name = fn
  pg._font_styles = fss
  pg._font_size = fsize
  pg._text_color = c
  pg._text_position = _new_text_position()
  pg._text_position.line_offset = 0
  pg._text_position.code_point_offset = 0
  pg._page_position = _new_page_position()
  pg._page_position.x_offset = 0.0  
  # x-offset in points of start of current line on page, or 0.0
  pg._page_position.y_offset = 0.0  
  # y-offset in points of start of current line on page
  pg._EOT_position = _new_page_position()
  pg._EOT_position.x_offset = 0.0
  pg._EOT_position.y_offset = 0.0
  # EOT position of empty text
  pg._line_list = [(0.0,"")]  # list of lines on page, as (x-offset as float, line as str)
  pg._back_map = [[(0,0)]]  # map from (page_line,glyph_offset) to (text_line, code_point_offset)
  return pg
  

def render_lines(t,curpos,pg,plo1,plo2):
  """
  pre:
    t = text which is being rendered
    curpos = cursor position in the text
    curpos start position in text <= curpos end position in text
    pg = Page to which text is to be rendered
    pg's back map must be set up for plo1
    plo1 = start page line offset for the render
    plo2 = end page line offset for the render
    if plo2 > last line offset of pg, pg will be rendered to end
    0 <= plo1 <= last line of text on page
    plo2 > 0
    plo2 > plo1
    
  post:
    the portion of text corresponding to the section of the page 
      between plo1 and the line before plo2
        has been rendered to pg
    curpos = cursor position in Text and Page
    any thin cursor has been rendered
    any fat cursor in the range plo1 <= cursor position < plo2 has been rendered

  pre:
    curpos._update_x_offset = True, iff _desired_x_offset is to be updated to the page x-offset

  post:
    curpos._update_x_offset = True
    curpos._desired_x_offset is undefined (-1.0)
    pg._page_position.x_offset = x-offset in points of start of plo2
    pg._page_position.y_offset = y-offset in points of start of plo2
    pg._line_list = list of (x-offset in points, line) for each line of page
    pg._back_map = list of lists of (text line offset, text code point offset)
                     for each glyph interstice in the page

  test:
    t = None
    t = ""
      curpos = None
      curpos = ((0,1),(0,0))
      curpos = ((0,0),(0,0))
        pg = None
        pg = valid page
          plo1 = None
          plo1 = -1
            plo2 = None
            plo2 = 0
          plo1 = 0
            plo2 = 0
          plo1 = 1
            plo2 = 0
            plo2 = 1 (should give plo1 out-of-range)
    t = text of three lines (one right-justified)
      plo1 = 1
        plo2 = 2
      plo1 = 0
        plo2 = TO_END
          cursor position in text is (0,0)
          cursor position in text is (0,0),(0,1) (fat(ish) cursor)
          cursor position in text is (0,4)
            curpos._update_x_offset = true
              curpos._update_x_offset = false
          cursor position in text is (0,5)
            curpos._update_x_offset = false
              curpos._update_x_offset = true
          cursor position in text is (0,8)
            curpos._update_x_offset = true
              curpos._update_x_offset = true
  """
  start_time = time.time()
  type_checking2_0.check_derivative(t,texting.Text)
  type_checking2_0.check_derivative(curpos,CursorPosition)
  type_checking2_0.check_derivative(pg,Page)
  type_checking2_0.check_identical(plo1,int)
  type_checking2_0.check_identical(plo2,int)
  if _less_than_text_position(curpos._end_text_position,curpos._start_text_position):
    raise Exception("Attempt to render lines with curpos._end_text_position < curpos._start_page_position")
  if plo1 < 0:
    raise Exception("Attempt to render lines with plo1 < 0")
  if plo2 <= 0:
    raise Exception("Attempt to render lines with plo2 <= 0")
  if plo2 <= plo1:
    raise Exception("Attempt to render lines with plo2 <= plo1")
  win = pg._my_window
  
  # wipe any thin cursor
  cursoring.wipe_cursor(win)
  
  # find start and end positions in text and page
  try:
    (stlo,stcpo) = text_position_of(plo1,0,pg)
  except IndexError:
    raise Exception("Attempt to render lines with plo1 > end of text on page")
  try:
    (etlo,etcpo) = text_position_of(plo2,0,pg)
  except IndexError:
    (etlo,etcpo) = (math.inf,math.inf)
  # set up pg to start of plo1
  pg._text_position.line_offset = stlo
  pg._text_position.code_point_offset = stcpo
  pg._page_position.x_offset = 0.0
  pg._page_position.y_offset = plo1 * pg._font_size
  pg._current_line = unicoding3_0.new_string()
  pg._line_width = 0.0
  
  # set up text cursor to equivalent position in text
  texting.set_cursor(t,stlo,stcpo)
  
  # check cursor
  _fix_cursor_position(curpos,pg)

  # set up initial alignment
  pg._current_alignment = texting.get_alignment(t)
  
  # read tokens until (etlo,etcpo) or <end> is reached
  eot = False
  while (texting.cursor_line_offset(t) != etlo \
    or texting.cursor_code_point_offset(t) != etcpo) \
    and not eot:
    tk = text_tokenizing.get_next_token(t)
    if unicoding3_0.equals(tk,unicoding3_0.string_of("<end>")):
      eot = True
    _render_token(tk,texting.get_alignment(t),curpos,pg)    
  curpos._update_x_offset = True
  render_thin_cursor(curpos,pg)
  end_time = time.time()
  #print("rl="+str(end_time - start_time))


def render_thin_cursor(curpos,pg):
  """ 
  pre:
    curpos = OnScreenCursorPositions of start and end of cursor respectively 
              in text and on page
    start text position <= end text position
    pg = Page on which cursor is to be rendered
  
  post:
    if the cursor is thin,
      if the cursor is totally in the text area of pg
        cursor has been (set up to be eventually) rendered on pg
      else
        an exception has been raised

  test:
    fat cursor
    thin cursor at page position (-1.0,-1.0)
    thin cursor at page position (0.0,0.0)
    thin cursor at page position (36.0,72.0)
      cursor just fits on text area
      cursor just overlaps end of text area
  """
  type_checking2_0.check_derivative(curpos,CursorPosition)
  type_checking2_0.check_derivative(pg,Page)
  if _less_than_text_position(curpos._end_text_position,curpos._start_text_position):
    raise Exception("Attempt to render cursor with start position > end position")
  if _equals_text_position(curpos._start_text_position,curpos._end_text_position):
    if curpos._start_page_position.y_offset < 0.0 or \
       curpos._start_page_position.y_offset + pg._font_size > pg._text_height:
         raise Exception("Attempt to render cursor at ("+ \
           str(curpos._start_page_position.x_offset)+","+ \
           str(curpos._start_page_position.y_offset)+ \
           ") which is out of range of text")
    else:
      cursoring.draw_cursor( \
        pg._my_window, \
        pg._font_size, \
        pg._horizontal_indent + curpos._start_page_position.x_offset, \
        pg._vertical_indent + curpos._start_page_position.y_offset, \
        coloring.BLACK)


def set_cursor_position_to_start(curpos):
  """ 
  pre:
    curpos = cursor position which is to be set to start of text
  
  post:
    curpos has been set to the start of the text, 
      and the start of the text on the page

  test:
    once thru with current text position set to (1,2)
  """
  type_checking2_0.check_derivative(curpos,CursorPosition)
  curpos._start_text_position.line_offset = 0
  curpos._start_text_position.code_point_offset = 0
  curpos._end_text_position.line_offset = 0
  curpos._end_text_position.code_point_offset = 0
  curpos._start_page_position.x_offset = 0.0
  curpos._start_page_position.y_offset = 0.0
  curpos._end_page_position.x_offset = 0.0
  curpos._end_page_position.y_offset = 0.0
  curpos._desired_x_offset =  - 1.0
  curpos._update_x_offset = True
  

def set_text_font_name(pg,fn):
  """ 
  pre:
    pg = Page whose font name is to be changed
    fn = font name in which pg's text is to be rendered from now on, as str
  
  post:
    pg's associated font name has been changed to fn

  test:
    once thru with new font name
  """
  type_checking2_0.check_derivative(pg,Page)
  type_checking2_0.check_derivative(fn,str)
  pg._font_name = fn
  

def set_text_font_size(pg,d):
  """ 
  pre:
    pg = Page whose font size is to be changed
    d = font size in which pg's text is to be rendered from now on
  
  post:
    pg's associated font size has been changed to d

  test:
    once thru with new font size
  """
  type_checking2_0.check_derivative(pg,Page)
  type_checking2_0.check_identical(d,float)
  pg._font_size = d
  

def set_thin_cursor_text_and_page(curpos,lo,cpo,xoff,yoff):
  """
  pre:
    curpos = present position of cursor
    lo = line offset of thin cursor in texting.Text
    cpo = code point offset of thin cursor in texting.Text
    xoff = x-offset of thin cursor on Page, relative to tlh corner of text on page
    yoff = y-offset of thin cursor on Page, relative to tlh corner of text on page
    
  post:
    curpos has been updated with the text and page positions
      of the thin cursor
  
  test:
    once thru
  """
  curpos._start_text_position.line_offset = lo
  curpos._start_text_position.code_point_offset = cpo
  curpos._end_text_position.line_offset = lo
  curpos._end_text_position.code_point_offset = cpo
  curpos._start_page_position.x_offset = xoff
  curpos._start_page_position.y_offset = yoff
  curpos._end_page_position.x_offset = xoff
  curpos._end_page_position.y_offset = yoff


def set_on_screen_to_text(t,curpos):
  """ 
  pre:
    t = text with internal cursor
    curpos = cursor position which is to be set to t's internal cursor
  
  post:
    curpos's text position has been set to t's internal cursor

  test:
    once thru
  """
  type_checking2_0.check_derivative(t,texting.Text)
  type_checking2_0.check_derivative(curpos,CursorPosition)
  curpos._start_text_position.line_offset = texting.cursor_line_offset(t)
  curpos._start_text_position.code_point_offset = texting.cursor_code_point_offset(t)
  curpos._end_text_position.line_offset = texting.cursor_line_offset(t)
  curpos._end_text_position.code_point_offset = texting.cursor_code_point_offset(t)
  

def set_text_to_on_screen_start(curpos,t):
  """ 
  pre:
    curpos = cursor position in text
    t = text whose internal cursor is to be set to 
          curpos's start position in the text
  
  post:
    t's internal cursor has been set to 
      curpos's start position in the text

  test:
    once thru
  """
  type_checking2_0.check_derivative(curpos,CursorPosition)
  type_checking2_0.check_derivative(t,texting.Text)
  texting.set_cursor(t,curpos._start_text_position.line_offset, \
                       curpos._start_text_position.code_point_offset)


def set_text_to_on_screen_end(curpos,t):
  """ 
  pre:
    curpos = cursor position in text
    t = text whose internal cursor is to be set to 
          curpos's end position in the text
  
  post:
    t's internal cursor has been set to 
      curpos's end position in the text

  test:
    once thru
  """
  type_checking2_0.check_derivative(curpos,CursorPosition)
  type_checking2_0.check_derivative(t,texting.Text)
  texting.set_cursor(t,curpos._end_text_position.line_offset, \
                       curpos._end_text_position.code_point_offset)


def start_page_line_of(curpos,pg):
  """
  pre:
    curpos = CursorPosition in page
    pg = Page in which cursor is positioned
 
  post:
    returns the start page line for a render_lines scan, 
      when the cursor has been moved forwards in the text
  
  test:
    once thru  
  """
  splo = int(curpos._start_page_position.y_offset/pg._font_size)
  return splo
  

def start_page_line_of_back(curpos,pg):
  """
  pre:
    curpos = CursorPosition in page
    pg = Page in which cursor is positioned
 
  post:
    returns the start page line for a render_lines scan, 
      when the cursor has been moved backwards in the text
  
  test:
    curpos at page line 2
    curpos at page line 1
    curpos at page line 0
  """
  splo = int(curpos._start_page_position.y_offset/pg._font_size)
  if splo > 0:
    splo -= 1
  return splo
  

def text_position_of(lo,go,pg):
  """
  pre:
    lo = line offset of cursor page position
    lo >= 0
    go = glyph offset of cursor page position
    go >= 0
    pg = Page in which cursor is positioned
    pg._back_map = map from page position to the equivalent text position
    
  post:
    returns text position equivalent to cursor page position, as an int pair
  
  test:
    lo non-integer
    lo = -1
    lo = 0
      go non-integer
      go = -1
      go = 0
        pg not a Page
        pg is a Page
    lo = 1
      go = 1
        page position exists
  """
  type_checking2_0.check_identical(lo,int)
  type_checking2_0.check_identical(go,int)
  type_checking2_0.check_derivative(pg,Page)
  if lo < 0:
    raise Exception("Attempt to find text position of negative page line offset")  
  if go < 0:
    raise Exception("Attempt to find text position of negative page glyph offset")
  (tlo,cpo) = pg._back_map[lo][go]
  return (tlo,cpo)
  

# private constants
# -----------------

_LIGHT_BLUE = coloring.new_color(0.9,0.9,1.0)
  

# -------------
# private types
# -------------

""" 
Note: all measurements on the page are done relative to the tlh corner of the text

Invariants:
  1. Negative values for _text_position or _page_position indicate an undefined position.
  2. Either the _text_position or the _page_position must be defined.
  3. If both positions are defined, the text position takes precedence.
"""
  
class _PagePosition:
  """  position in the page as (x-offset:float, y_offset:float)"""
  pass
  

class _TextPosition:
  """ position in the text as (line offset:int, code point offset:int)"""
  pass
  

# private procedures
# ------------------

def _de_escape(s,curpos,pg):
  """ 
  pre:
    s = unicoding3_0.String of escaped unicode code points which is to be 
          de-escaped and appended to pg._current_line
    curpos._start_text_position = curpos's start text position
    curpos._start_page_position = curpos's start page position
    curpos._end_text_position = curpos's end text position
    curpos._end_page_position = curpos's end page position
    curpos._desired_x_offset = desired x-offset of curpos, or -1.0 if undefined
    pg = page which is being rendered to
    pg._font_name = font name for this render, as a str
    pg._font_styles = font_styling.FontStyles for this render
    pg._font_size = font size for this render, as a float
    pg._text_position = current render position in text, as (line_offset, code_point_offset)
    pg._page_position.x_offset = x-offset of start of current line in page, 
                                   or 0.0
    pg._page_position.y_offset = y-offset of start of current line in page
    pg._current_line = unicoding3_0.String which is to be appended to
    pg._my_window = windowing.Window for this render
    pg._back_map = map from (page_line,glyph_offset) to 
      (text_line,code_point_offset) for the text rendered so far
  
  post:
    s has been de-escaped and appended to pg._current_line
    for each de-escaped code point (glyph) of s:
      an entry has been added to the pg._back_map,
        giving the text position for the current glyph interstice
      the cursor start and end points have been updated as necessary

  test:
    font = "Courier New"
      font size = 12
        pg._text_position = (0,0)
          pg._page_position = (0.0,0.0)
            pg._current_line = ""
              pg._back_map = [[(0,0)]]
                s = ""
                  curpos and curpos at (0,0) (0.0,0.0)  
                    (check cursor positions, check _back_map)
        pg._text_position = (1,2)
          pg._page_position = (14.0,12.0)
            pg._current_line = "ab"
              pg._back_map = [[(0,0)],[(0,0)]]
                s = "1&lt;2"
                  curpos at (1,3) (-1.0,-1.0)
                    curpos at (1,7) (-1.0,-1.0)
                      (check cursor page positions, check _back_map)
  """
  i = 0
  while i < unicoding3_0.length_of(s):
    old_offset = i
    (code_point,i) = string_parsing.parse_html_character(s,i)
    width_in_code_points = i - old_offset
    pg._text_position.code_point_offset += width_in_code_points
    _render_glyph(code_point,curpos,pg)
    # update _back_map with current text position
    _update_back_map(pg)


def _ensure_sublist_exists(l,i):
  """
  pre:
    l = list which is to be modified, if necessary
    i = offset in l of sub-list
    
  post:
    l has been updated with an empty sub-list, if necessary,
      otherwise l is left unchanged
     
  test:
  
    
  """
  if i < len(l):
    pass
  elif i == len(l):
    l.append([])
  else:
    raise Exception("Attempt to create sublist in l with out-of-range index. i="+str(i)+", len(l)="+str(len(l)))
  
    
def _equals_float(d1,d2):
  """ 
  pre:
    d1 = first floating-point number
    d2 = second floating-point number
    
  post:
    returns true iff the two floating-point numbers are within 0.01 of each other

  test:
    d1 = 123.4
      d2 = 123.3    
      d2 = 123.401    
      d2 = 123.5    
  """
  return abs(d1 - d2) < 0.01
  

def _equals_text_position(tp1,tp2):
  """ 
  pre:
    tp1 = first position in text
    tp2 = second position in text
  
  post:
    returns true iff the two positions in the text are equal

  test:
    tp1 = (1,2)
      tp2 = (0,2)
      tp2 = (2,2)
      tp2 = (1,1)
      tp2 = (1,3)
      tp2 = (1,2)     
  """
  return (tp1.line_offset == tp2.line_offset) and (tp1.code_point_offset == tp2.code_point_offset)
  

"""
def _find_text_and_page_offset(curpos,lm,lso):
  """ """ 
  pre:
    curpos._desired_x_offset = desired x-offset of cursor on page
    curpos._page_position.y_offset = y-offset of cursor on page
    curpos._text_position.line_offset = offset of line being rendered
    lm = map of glyph-offsets to text code point offsets and relative page x-offsets
          for the current line on the page
    lso = start-offset of current line on the page
  
  post:
    curpos._text_position = position of cursor in internal text
    curpos._page_position = position of cursor on page

  test:
    lso = 20.0
      curpos._text_position.line_offset = -1;
      curpos._current_text_position.line_offset = 0;
        map with single entry (0 -> (0,0.0))
          curpos._desired_x_offset = 19.9
          curpos._desired_x_offset = 20.1
        map with entries (0 -> (0,0.0)), (1 -> (12,34.0)),(2 -> (13,35.0)), (3 -> (14,36.0))
          curpos._desired_x_offset = 19.9
          curpos._desired_x_offset = 54.5
          curpos._desired_x_offset = 54.6
          curpos._desired_x_offset = 56.1      
  """ """
  # relativize the desired position to the current start offset
  rdxo = curpos._desired_x_offset - lso
  # find the nearest page x-offset
  minimum_page_distance = float("infinity")
  i = 0
  while i < len(lm.map):
    ptp = lm.map.get(i)
    difference = abs(rdxo - ptp.page_relative_x_offset)
    if difference < minimum_page_distance:
      # we have a minimum (so far)
      minimum_page_distance = difference
      curpos._page_position.x_offset = ptp.page_relative_x_offset
      curpos._text_position.code_point_offset = ptp.text_code_point_offset    
    i += 1
  # set the line offset
  curpos._text_position.line_offset = curpos._current_text_position.line_offset
  # derelativize the x-offset
  curpos._page_position.x_offset += lso
"""
  
def _fix_cursor_position(curpos,pg):
  """ 
  pre:
    curpos.start_text_position is the start text position of the cursor,
      or (-1,-1) if undef
    curpos.end_text_position is the end text position of the cursor,
      or (-1,-1) if undef
    pg._text_position is the current text position of the render
    pg._page_position.y_offset =  y-offset of the current line on the page,
                                    in points as float
    pg._current_line = current line of render, as a unicoding3_0.String
    pg._font_name = font name for this render, as a str
    pg._font_styles = font_styling.FontStyles for this render
    pg._font_size = font size for this render, as a float
    pg._my_window = windowing.Window for this render
    
  post:
    if curpos's start text position is defined,
      if curpos's start text position corresponds to the current text position
        curpos._start_page_position.x_offset = start x-offset relative to start of line
        curpos._start_page_position.y_offset = start y-offset relative to start of page
        if the cursor is thin,
          curpos._end_page_position.x_offset = end x-offset relative to start of line
          curpos._end_page_position.y_offset = end y-offset relative to start of page
        else (cursor is fat),
          curpos._start_x_offset = start x-offset of cursor rectangle,
                                 relative to start of line
          curpos._in_fat_cursor = True
    if curpos's end text position is defined,
      if curpos's end text position corresponds to the current text position
        curpos._end_page_position.x_offset = x-offset relative to start of line
        curpos._end_page_position.y_offset = y-offset relative to start of page
        if cursor is fat,
          curpos._start_x_offset = start x-offset of cursor rectangle,
                                 relative to start of line
          curpos._end_x_offset = end x-offset of cursor rectangle,
                               relative to start of line
          curpos._in_fat_cursor = False
          curpos._cursor_rectangle_pending = True

  test:
    curpos._start_text_position = (-1,-1)
    curpos._start_text position = (0,1)
        pg._current_line = unicoding3_0.string_of("f")
          pg._font_name = "Courier New"
            pg._font_styles = {}
              pg._font_size = 12.0
                pg._text_position = (0,2)
                pg._text_position = (0,1)
                  pg._page_position.x_offset = 7.2
                    pg._page_position.y_offset = 0.0
                      curpos._end_text_position = (0,1)
                      curpos._end_text_position = (0,2)
                pg._text_position = (0,2)
                  pg._page_position.x_offset = 14.4
                    pg._page_position.y_offset = 0.0
                      curpos._end_text_position = (0,2)
    curpos._end_text_position = (-1,-1)
    curpos._end_text position = (0,3)
      pg._font_name = "Courier New"
        pg._font_styles = {}
          pg._font_size = 12.0
            pg._text_position = (0,2)
              pg._current_line = unicoding3_0.string_of("ff")
            pg._text_position = (0,3)
              pg._current_line = unicoding3_0.string_of("fff")
                pg._page_position.x_offset = 21.6
                  pg._page_position.y_offset = 0.0
  """
  if curpos._start_text_position.line_offset >= 0:
    if _equals_text_position(curpos._start_text_position,pg._text_position):
      line_as_str = unicoding3_0.python_string_of(pg._current_line)
      line_width_in_points = writing.width_in_points_of(pg._my_window,line_as_str,pg._font_name,pg._font_styles,pg._font_size)
      curpos._start_page_position.x_offset = line_width_in_points
      curpos._start_page_position.y_offset = pg._page_position.y_offset
      if _equals_text_position(curpos._start_text_position,curpos._end_text_position):
        # thin cursor
        curpos._end_page_position.x_offset = line_width_in_points
        curpos._end_page_position.y_offset = pg._page_position.y_offset
        curpos._in_fat_cursor = False
        curpos._cursor_rectangle_pending = False
      else:  # fat cursor
        curpos._start_x_offset = line_width_in_points
        curpos._in_fat_cursor = True
        curpos._cursor_rectangle_pending = False
  if curpos._end_text_position.line_offset >= 0:
    if _equals_text_position(curpos._end_text_position,pg._text_position):
      line_as_str = unicoding3_0.python_string_of(pg._current_line)
      line_width_in_points = writing.width_in_points_of(pg._my_window,line_as_str,pg._font_name,pg._font_styles,pg._font_size)
      curpos._end_page_position.x_offset = line_width_in_points
      curpos._end_page_position.y_offset = pg._page_position.y_offset
      if curpos._in_fat_cursor:
        curpos._end_x_offset = line_width_in_points
        curpos._in_fat_cursor = False
        curpos._cursor_rectangle_pending = True


def _less_than_text_position(tp1,tp2):
  """ 
  pre:
    tp1 = first position in text
    tp2 = second position in text
  
  post:
    returns true iff curpos < curpos

  test:
    tp1 = (1,2)
      tp2 = (0,2)
      tp2 = (2,2)
      tp2 = (1,1)
      tp2 = (1,3)
      tp2 = (1,2)     
  """
  return (tp1.line_offset < tp2.line_offset) or \
         (tp1.line_offset == tp2.line_offset) and \
         (tp1.code_point_offset < tp2.code_point_offset)
  

def _new_page_position():
  """
  pre:
    true
    
  post:
    new _PagePosition variable has been returned,
      with undefined position
  
  note: 
    the PagePosition gives the position of the cursor in points 
      wrt the tlh corner of the text on the page
      
  test:
    once thru
  """
  pp = _PagePosition()
  pp.x_offset =  -1.00
  pp.y_offset =  -1.00
  return pp


def _new_text_position():
  """
  pre:
    true
    
  post:
    new _TextPosition variable has been returned,
      with undefined position
      
  note:
    the TextPosition gives the position of the cursor in the text as 
      (text_line_offset, code_point_offset)
    
  test:
    once thru
  """
  tp = _TextPosition()
  tp.line_offset =  -1
  tp.code_point_offset =  -1
  return tp


def _render_cursor_rectangle(x1,x2,pg):
  """
  pre:
    x1,x2 = start and end x-offsets of cursor rectangle in current line, as floats
    x1 < x2
    pg = Page on which curpos is to be rendered
    
  post:
    fat cursor rectangle has been rendered as required
    
  test:
    once thru
  """
  win = pg._my_window
  x = x1 + pg._horizontal_indent
  y = pg._vertical_indent + pg._page_position.y_offset
  w = x2 - x1
  h = pg._font_size
  c = _LIGHT_BLUE
  painting.paint_rectangle(win,x,y,w,h,c)

"""
def _render_cursor_rectangle(curpos,pg):
  """ """
  pre:
    curpos = start and end positions of cursor in text and page
    curpos < curpos
    pg = Page on which curpos is to be rendered
    curpos._in_fat_cursor = True
    curpos._in_fat_cursor = True
    
  post:
    fat cursor rectangle has been rendered as required
    
  test:
    once thru
  """ """
  win = pg._my_window
  x = pg._horizontal_indent + curpos._start_page_position.x_offset
  y = pg._vertical_indent + curpos._start_page_position.y_offset
  w = curpos._end_page_position.x_offset - curpos._start_page_position.x_offset
  h = pg._font_size
  c = _LIGHT_BLUE
  painting.paint_rectangle(win,x,y,w,h,c)
"""

def _render_glyph(cp,curpos,pg):
  """
  pre:
    cp = Unicode code point which is to be rendered
    curpos = start and end positions of cursor
            (Page positions may be undefined)
    pg = Page on which cp is to be rendered
    pg._text_position = position in text just after glyph to be rendered
    pg._page_position.y_offset = current y-offset of render
    
  post:
    cp has been, or shortly will be, rendered on pg
    the cursor has been rendered as necessary
    
  tests:
    once thru with cursor positioned after glyph
  """
  # append cp to current line
  unicoding3_0.append(pg._current_line,cp)
  
  # check cursor
  _fix_cursor_position(curpos,pg)
  

def _render_line(pg,curpos):
  """ 
  pre:
    pg = Page for this render
    pg._current_line = line to be rendered
    pg._line_width = width of current_line in points
    pg._text_width = width of text in points
    pg._current_alignment = alignment of line to be rendered
    pg._line_list = list of lines rendered, as (x_offset as float, line as str)
    pg._text_position.line_offset = offset of line being rendered in text
    pg._text_position.code_point_offset = offset of current code point in text line
    pg._page_position.y_offset = y-offset of line being rendered in page
    pg._my_window = windowing.Window for this render
    curpos._start_x_offset = start position of cursor rectangle
    curpos._end_x_offset = end position of cursor rectangle
    curpos._in_fat_cursor = True, iff render of fat cursor has been started
    
  post:
    if current line is on the page,
      pg._current_line' has been rendered
      (start x_offset,pg._current_line') has been added to the line_map
      if a cursor position was on line,
        its x-offset has been derelativized
    pg.page_position.x_offset = 0.0
    pg.page_position.y_offset = y-offset for next line
    pg._current_line = empty string
    pg._line_width = 0.0
    if the fat cursor is totally or partially on this line,
      it has been rendered

  test:
    line "fred"
      current_alignment = BEGIN
        pg._text_position = (0,0)
      current_alignment = END
        pg._text_position = (1,0)
      current_alignment = MIDDLE
        pg._text_position = (1,4)
  """
  # render the current line
  _try_line(pg,curpos)
  # set up the next line
  pg._page_position.x_offset = 0.0
  pg._page_position.y_offset += pg._font_size
  pg._current_line = unicoding3_0.new_string()
  pg._line_width = 0.0
  

def _render_token(tk,a,curpos,pg):
  """ 
  pre:
    tk = token to be rendered, as 
    a = alignment of text following tk
    curpos._start_text_position = curpos's start text position
    curpos._start_page_position = curpos's start page position or (-1.0,-1.0) 
                                if undefined
    curpos._desired_x_offset = desired x-offset of curpos, or -1 if undefined
    curpos._end_text_position = curpos's end text position
    curpos._end_page_position = curpos's end page position or (-1.0,-1.0) if undefined
    curpos._desired_x_offset = desired x-offset of curpos, or -1 if undefined
    curpos._in_fat_cursor = True, iff current position is inside a fat cursor
    curpos._cursor_rectangle_pending = True, iff a cursor rectangle has been 
                                     determined
    curpos._start_x_offset = cursor rectangle start x-offset relative to start of 
      line, if curpos._in_fat_cursor or curpos._cursor_rectangle_pending
    curpos._end_x_offset = cursor rectangle end x_offset relative to start of line,
      if curpos._cursor_rectangle_pending
    pg = Page to which this token is to be rendered
    pg._text_position = current position of render in text (line, code point)
    pg._line_width = width of current line in points
    pg._page_position.x_offset = 
      x-offset of start of current line in page, or 0.0.
    pg._page_position.y_offset = y-offset of start of current line in page
    pg._current_line = current line being rendered, as unicoding3_0.String
    pg._EOT_position.x_offset = x-offset of start of last line of text
    pg._EOT_position.y_offset = y-offset of start of last line of text
    pg._my_window = window associated with this page
    pg._back_map = map from (page_line,glyph_offset) to 
      (text_line,code_point_offset) for the text rendered so far
  
  post:
    curpos._start_text_position = curpos's start text position
    curpos._start_page_position = curpos's start page position 
                                or (-1.0,-1.0) if undefined
    curpos._end_text_position = curpos's end text position
    curpos._end_page_position = curpos's end page position or (-1.0,-1.0) if undefined
    curpos._desired_x_offset = desired x-offset of curpos, or -1 if undefined
    curpos._in_fat_cursor = True, iff current position is inside a fat cursor
    curpos._cursor_rectangle_pending = True, iff a cursor rectangle has been 
                                     determined
    curpos._start_x_offset = cursor rectangle start x-offset 
                               relative to start of line, 
                                 if curpos._in_fat_cursor or   
                                   curpos._cursor_rectangle_pending
    curpos._end_x_offset = cursor rectangle end x_offset 
                             relative to start of line,
                               if curpos._cursor_rectangle_pending
    pg._text_position = current position of render in text (line, code point)
    pg._page_position.y_offset = y_offset of current position of render in page
    pg._EOT_position.x_offset = end-of-text position x-offset
    pg._EOT_position.y_offset = end-of-text position y-offset
    pg._line_width = width of current line in points
    pg._back_map = map from (page_line,glyph_offset) to 
      (text_line,code_point_offset) for the text rendered so far
    if tk is an <end> token,
      the current line has been rendered
      curpos._start_page_position = curpos's start page position
      curpos._end_page_position = curpos's end page position
    if tk is a Separator token,
      the width in points of the line with a space appended is calculated
      if overflow has occurred,
        the current line has been rendered
        a new empty line has been started
      else
        the space has been appended to the line
    if tk is a Newline token,
      the current line has been rendered
      the alignment has been set to a
    if tk is a Word token,
      the width in points of the line with the word appended is calculated
      if overflow has occurred,
        any trailing space has been removed from the current line
        any pending cursor rectangle has been shortened
        the current line has been rendered
        a new line has been started with the word
      else
        the word has been appended to the line
    
  test:
    pg._current_alignment = BEGIN
      "<finish>"
      "<end>"
      "the"
        curpos = curpos = 0,0
        curpos = 0,1
          curpos = 0,3
      "the cat"
        curpos = 0,4
          curpos = 0,7
      "the cat sat on the mat"<end> (check EOT_position)
        "\n"
          a = MIDDLE
            "\n"
              a = END
                cursor at 0,18
                cursor at 0,19
                cursor at 0,22
                cursor at 1,0
                cursor at 2,0
      "the dogs sat on the"
        curpos = curpos = 0,0
      "the dogs sat on the "
        curpos = curpos = 0,0
  """
  try:
    # first code point of token gives sort
    cp0 = unicoding3_0.code_point_at(tk,0)
    # Tag
    if cp0 == ord('<'):
      la = looking_ahead.lookahead_of_string(tk)
      label = html_parsing.parse_tag(la)
      if unicoding3_0.equals(label,unicoding3_0.string_of("end")):
        _try_line(pg,curpos)
        pg._EOT_position.x_offset = pg._page_position.x_offset
        pg._EOT_position.y_offset = pg._page_position.y_offset
      else:
        raise Exception("unexpected tag label when rendering text:" + unicoding3_0.python_string_of(label))
        
    # Separator
    elif cp0 == ord(' '):
      pg._text_position.code_point_offset += 1
      space_width = writing.width_in_points_of(pg._my_window, ' ', pg._font_name, pg._font_styles, pg._font_size)
      pg._line_width += space_width
      # check for overflow of text width
      if pg._line_width > pg._text_width:
        pg._line_width -= space_width
        # render line and start a new line
        _render_line(pg,curpos)
        # update the _back_map
        _update_back_map(pg)
        # fix the cursor position if at start of new line
        _fix_cursor_position(curpos,pg)
      else:
        _render_glyph(ord(' '),curpos,pg)
        _update_back_map(pg)
        
    #Newline
    elif cp0 == ord('\n'):
      _render_line(pg,curpos)
      pg._current_alignment = a
      pg._text_position.line_offset += 1
      pg._text_position.code_point_offset = 0
      _update_back_map(pg)
      if _equals_text_position(curpos._end_text_position, pg._text_position):
        # render a rectangle of width 0.0 (i.e. do nothing)
        curpos._in_fat_cursor = False
      if curpos._in_fat_cursor:
        # reset cursor start text and page position
        curpos._start_text_position.line_offset = pg._text_position.line_offset
        curpos._start_text_position.code_point_offset = pg._text_position.code_point_offset
        curpos._start_page_position.x_offset = pg._page_position.x_offset
        curpos._start_page_position.y_offset = pg._page_position.y_offset
        # set cursor end page position undef
        curpos._end_page_position.x_offset = -1.0
        curpos._end_page_position.y_offset = -1.0
      else:  # thin cursor
        _fix_cursor_position(curpos,pg)
    
    # Word
    else:
      # remember the current page line width in points
      old_page_line_width = pg._line_width
      word_width = writing.width_in_points_of(pg._my_window, unicoding3_0.python_string_of(tk), pg._font_name, pg._font_styles, pg._font_size)
      pg._line_width += word_width
      # check for overflow of text width
      if pg._line_width > pg._text_width:
        # reset the line width
        pg._line_width = old_page_line_width
        # remove any trailing space
        last_pos = unicoding3_0.length_of(pg._current_line)-1
        if last_pos >= 0:
         if unicoding3_0.code_point_at(pg._current_line,last_pos) == ord(' '):
            unicoding3_0.remove(pg._current_line,last_pos)
            space_width = writing.width_in_points_of(pg._my_window, ' ', pg._font_name, pg._font_styles, pg._font_size)
            pg._line_width -= space_width
            if curpos._in_fat_cursor:
              if curpos._start_x_offset > pg._line_width:
                curpos._in_fat_cursor = False
            if curpos._cursor_rectangle_pending:
              if curpos._end_x_offset > pg._line_width:
                curpos._end_x_offset = pg._line_width
        # render the line without the current word
        _render_line(pg,curpos)
        # update the _back_map
        _update_back_map(pg)
        # fix the cursor position if at start of new line
        _fix_cursor_position(curpos,pg)
        # translate the word to the new line
        _de_escape(tk,curpos,pg)
        pg._line_width = word_width
      else:
        _de_escape(tk,curpos,pg)
  except Exception as inst:
    raise Exception("Unexpected exception when rendering token:"+str(inst))
    
  
def _try_line(pg,curpos):
  """ 
  pre:
    pg = Page to be rendered
    pg._current_line = line of page to be rendered
    pg._page_position.y_offset = current y offset of render on page, 
      in points relative to the tlh corner of the text, as float 
    pg._line_width = width of current line in points as float
    pg._text_width = maximum width of text on page in points as float
    pg._current_alignment = aligment of current line, as texting.Alignment
    pg._font_size = point size of font for this render, as float
    pg._text_height = height of text area on page, in points, as flaot 
    pg._line_list = list of lines rendered on page, as pair (x-offset as float, line as str)
    curpos = position of cursor, in text and page
    curpos._in_fat_cursor = 
      True, iff start position of fat cursor on page has been set,
        relative to the start of line, but end position has not  been found
    curpos._cursor_rectangle_pending = 
      True, iff final cursor rectangle has been created,
        relative to the start of line, but has not been rendered
    if curpos._in_fat_cursor,
      curpos._start_x_offset = start x-offset of cursor rectangle, 
                                 relative to start of line
    if curpos._cursor_rectangle_pending,
      curpos._start_x_offset = start x-offset of cursor rectangle, 
                                 relative to start of line
      curpos._end_x_offset = end x-offset of cursor rectangle,
                               relative to start of line

  post:
    if the current line is above the bottom of the text area,
      the current line has been rendered on the page
      the current line has been added to the line list
      if curpos._cursor_rectangle_pending,
        cursor rectangle has been aligned to the current line
        cursor rectangle has been queued for rendering
      if in_fat_cursor,
        cursor rectangle has been extended to the end of this line,
        cursor rectangle has been aligned to the current line
        non-final cursor rectangle has been queued for rendering
      if cursor is thin,
        the cursor has been aligned to the current line
      curpos._in_fat_cursor = 
        True, iff start position of fat cursor on page has been set,
          relative to the start of line, but end position has not  been found
      curpos._cursor_rectangle_pending = False
      pg._page_position.x_offset = x-offset of end of rendered line,
                                     relative to left-hand edge of text
      pg._page_position.y_offset = y-offset of end of rendered line,
                                     relative to top of text

  test:
    line at (0.0,0.0), cursor at ((0.0,0.0),(0.0,0.0))
    line just overlaps end of text area      
    line just fits on text area (check line list, pg._page_position)
      cursor start at start of previous line
        cursor end at start of previous line
      cursor start at start of line
        cursor end at start of line
      cursor start at start of line
        cursor end at start of line + 7.2
      cursor start at start of previous line
        cursor end at start of previous line + 7.2
    line penultimate of page
      cursor start at start of previous line
        cursor ends on this line
          curpos._in_fat_cursor False
            curpos._cursor_rectangle_pending True
              curpos._start x_offset  = start of this line
                curpos._end_x_offset = end point of cursor on this line
        cursor ends on next line
          curpos._in_fat_cursor True
            curpos._cursor_rectangle_pending False
              curpos._start x_offset = start of this line
      cursor start in this line
        cursor ends on this line, but is fat
          curpos._in_fat_cursor False
            curpos._cursor_rectangle_pending True
              curpos._start x_offset = start offset of cursor in this line
                curpos._end_x_offset = end offset of cursor in this line
  """
  if pg._page_position.y_offset + pg._font_size <= pg._text_height:
    pg._page_position.x_offset = page_laying_out.x_offset_of_line(pg._line_width,pg._text_width,pg._current_alignment)
    line_as_str = unicoding3_0.python_string_of(pg._current_line)
    end_x_offset = \
      writing.write_string(pg._my_window, \
                           line_as_str, \
                           pg._font_name, \
                           pg._font_styles, \
                           pg._font_size, \
                           pg._horizontal_indent + pg._page_position.x_offset, \
                           pg._vertical_indent + pg._page_position.y_offset, \
                           pg._text_color) - pg._horizontal_indent
    page_line_offset = math.floor(pg._page_position.y_offset/pg._font_size)
    _update_list(pg._line_list,page_line_offset,(pg._page_position.x_offset,line_as_str))
    if curpos._cursor_rectangle_pending:
      # align the rectangle
      ax1 = curpos._start_x_offset + pg._page_position.x_offset
      ax2 = curpos._end_x_offset + pg._page_position.x_offset
      _render_cursor_rectangle(ax1,ax2,pg)
      curpos._cursor_rectangle_pending = False
    if curpos._in_fat_cursor:
      y_offset = pg._page_position.y_offset
      ax1 = curpos._start_x_offset + pg._page_position.x_offset
      # this ensures rectangle is aligned to the text line
      ax2 = pg._page_position.x_offset + pg._line_width
      # render the non-final cursor rectangle
      _render_cursor_rectangle(ax1,ax2,pg)
      # set up the cursor rectangle for the next line
      curpos._start_x_offset = 0.0
    else:  # align thin cursor
      if _equals_float(curpos._start_page_position.y_offset,pg._page_position.y_offset):
        curpos._start_page_position.x_offset += pg._page_position.x_offset
      if _equals_float(curpos._end_page_position.y_offset,pg._page_position.y_offset):
        curpos._end_page_position.x_offset += pg._page_position.x_offset
    pg._page_position.x_offset = end_x_offset


def _update_back_map(pg):
  """
  pre:
    pg = page which is being rendered to
    pg._back_map = back map from Page to Text which is being updated
    pg._font_size = font size for this render, as a float
    pg._text_position = current render position in text, as (line_offset, code_point_offset)
    pg._page_position.y_offset = y-offset of start of current line in page
    pg._current_line = unicoding3_0.String representing current line of page
    
  post:
    pg._back_map has been updated with an entry for the glyph interstice
      at the end of the current line
      
  test:
    test:
      pg._font_size = 12.0
        pg._current_line = "a"
          pg._text_position = (0,1)
            pg._page_position.y_offset = 0.0
        pg._current_line = ""
          pg._text_position = (1,0)
            pg._page_position.y_offset = 12.0
        pg._current_line = "b"
          pg._text_position = (1,1)
            pg._page_position.y_offset = 12.0
  """
  page_line_offset = math.floor(pg._page_position.y_offset/pg._font_size)
  _ensure_sublist_exists(pg._back_map,page_line_offset)
  _update_list(pg._back_map[page_line_offset],
     unicoding3_0.length_of(pg._current_line),
     (pg._text_position.line_offset,pg._text_position.code_point_offset))


def _update_list(l,i,v):
  """
  pre:
    l = list to be updated
    i = position in list to be updated (0 <= i <= len(l))
    v = value to be inserted in list at position i
    
  post:
    l has been updated with v at position i
    l may have been extended by one element
    
  test:
    l is empty
      i = 1
      i = 0
        v= 123
      i = 1
        v = 456
      i = 0
        v = 789
  """
  if i < len(l):
    l[i] = v
  elif i == len(l):
    l.append(v)
  else:
    raise Exception("Attempt to update list with out-of-range index. i="+str(i)+", len(l)="+str(len(l)))



# Contractor which converts text to tokens.

# author R.N.Bosworth

# version 8 Mar 23  10:56

from guibits1_0 import type_checking2_0, unicoding3_0
from . import code_point_testing, texting

""" 

Copyright (C) 2018,2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""
""" 
  Text Grammar version 4 Feb 2019   09:29
  
  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::==  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive.
  
  Text  ::==  OptionalSpaces TextLine TextLine* END_OF_TEXT
  
  TextLine  ::==  Element* NewLine
  
  Element  ::==  Word | Separator

  Word  ::==  NonSpaceChar NonSpaceChar*
  
  NonSpaceChar is any HtmlCharacter except LF, SPACE, TAB, FF or CR
  
  Separator  ::==  SPACE (U+0020)
  
  Newline  ::==  LF (U+000A)
                                                               
  Note 1: Rule for OptionalSpaces is given in HTMLCSSGrammar.  Rule for HtmlCharacter is given in HTMLGrammar.
  
  Note 2: An Internal Paragraph is a series of TextLines terminated by a blank line or END_OF_TEXT.  Every TextLine of an Internal Paragraph has the same alignment.     
"""
""" 
  Token Grammar version 11 Feb 2019   14:43

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::==  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive.
    

  Text  ::==  Token* EndTag
  
  Token  ::==  Word | Separator | NewLine | InternalTag
  
  EndTag  ::==  < e n d >
  
  Word  ::==  NonSpaceChar NonSpaceChar*
  
  NonSpaceChar is any HtmlCharacter except LF, SPACE, TAB, FF or CR
  
  Separator  ::==  SPACE (U+0020)
  
  Newline  ::==  LF (U+000A)
  
  InternalTag  ::==   (this will be extended later)
  
  Note: Rule for HtmlCharacter is given in HTMLGrammar.
"""


# exposed procedures
# ------------------

def get_next_token(t):
  """ 
  pre:
   t = texting.Text that is being parsed to tokens

  post:
    the next token of t has been returned
    t's cursor is just beyond the returned token 

  syntax:
    Text  ::=  Token* EndTag
    
    Token  ::=  Word | Separator | InternalTag
    
    EndPageTag  ::=  < e n d >

  test:
    t = None
    t = "<i> fred"
      call get_next_token 5 times
  """
  type_checking2_0.check_derivative(t,texting.Text)
  if texting.current_code_point(t) == texting.END_OF_TEXT:
    return unicoding3_0.string_of("<end>")
  else:
    return _parse_token(t)
    

# private procedures
# ------------------

def _parse_token(t):
  """ 
  pre:
    t = texting.Text which is being parsed
    
  post:
    the next token of t has been returned
    t's cursor is just beyond the returned token 

  syntax:
    Token  ::==  Word | Separator | NewLine | InternalTag
  
    Word  ::==  NonSpaceChar NonSpaceChar*
    
    NonSpaceChar is any HtmlCharacter except LF, SPACE, TAB, FF or CR
    
    Separator  ::==  SPACE (U+0020)
    
    Newline  ::==  LF (U+000A)
    
    InternalTag  ::==   (this will be extended later)
      
  test:
    text = " fred<b>x\nbert"
      call _parse_token 8 times
  """
  s = unicoding3_0.new_string()
  if texting.current_code_point(t) == ord('\n'):
    texting.advance(t)
    unicoding3_0.append(s,ord('\n'))
  elif code_point_testing.is_space_char(texting.current_code_point(t)):
    texting.advance(t)
    unicoding3_0.append(s,ord(' '))
  elif texting.current_code_point(t) == ord('<'):
    texting.advance(t)
    tag_label = unicoding3_0.new_string()
    while texting.current_code_point(t) != ord('>'):
      unicoding3_0.append(tag_label,texting.current_code_point(t))
      texting.advance(t)
    texting.advance(t)
    unicoding3_0.append(s,ord('<'))
    unicoding3_0.append_a_copy(s,tag_label)
    unicoding3_0.append(s,ord('>'))
  else:  # must be NonSpaceChar - start of Word
    unicoding3_0.append(s,texting.current_code_point(t))
    texting.advance(t)
    while texting.current_code_point(t) != ord('<') and ( not code_point_testing.is_space_char(texting.current_code_point(t)) and texting.current_code_point(t) != texting.END_OF_TEXT):
      unicoding3_0.append(s,texting.current_code_point(t))
      texting.advance(t)
  return s



from guibits1_0 import type_checking2_0
from . import unicode_io
from guibits1_0 import unicoding3_0

# author R.N.Bosworth

# version 2 Mar 2023  14:51
""" 
Contractor for emitting html elements.

Copyright (C) 2017,2018,2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.


"""
""" 
  HTML Grammar version 1 Dec 2018   15:05

  This grammar defines the basic HTML constructs common to any HTML file.

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::==  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive, unlike HTML5.
  
  
  StartTag(Name)  ::==  < Name OptionalSpaces >

  EndTag(Name)  ::==  < / Name OptionalSpaces >

  HtmlCharacter  ::==  ( & a m p ;
                       | & g t ;
                       | & l t ;
                       | any Unicode code point except '&' (U+26), '>' (U+3E) and '<' (U+3C)

Note: the rules for OptionalSpaces, SpaceChar, Separator, Name and NameChar are in the HTML CSS Grammar, as they are common to both HTML and CSS
"""

# exposed procedures
# ------------------

def emit_end_tag(s,w):
  """ 
  pre:
    s = name of end tag, as lower-case Unicode code points
    w = Unicode stream to which end tag is to be emitted
    
  post:
    end tag has been written to w

  syntax:
    EndTag(Name)  ::=  < / Name OptionalSpaces >

  test:
    s = None
    s = vaid non-empty string
      w = None
      w = valid Writer
  """
  type_checking2_0.check_derivative(s,unicoding3_0.String)
  type_checking2_0.check_derivative(w,unicode_io.Writer)
  unicode_io.write(ord('<'),w)
  unicode_io.write(ord('/'),w)
  i = 0
  while i < unicoding3_0.length_of(s):
    unicode_io.write(unicoding3_0.code_point_at(s,i),w)
    i += 1
  unicode_io.write(ord('>'),w)
  

def emit_indent(w):
  """ 
  pre:
    w = Writer on which indent is to be emitted
    
  post:
   indent has been written to w

  test:
    w = None
    w = valid Writer
  """
  type_checking2_0.check_derivative(w,unicode_io.Writer)
  unicode_io.write(ord(' '),w)
  unicode_io.write(ord(' '),w)
  

def emit_new_line(w):
  """ 
  pre:
    w = Unicode stream to which new line string is to be emitted
    
  post:
    new line string has been written to w

  test:
    w = None
    w = valid Writer
      unix
      windows
      default (unix)
  """
  type_checking2_0.check_derivative(w,unicode_io.Writer)
  emit_string(_LINE_SEPARATOR,w)
  

_LINE_SEPARATOR = unicoding3_0.string_of("\n") # default to Unix


def emit_string(s,w):
  """ 
  pre:
    s = string to be emitted
    w = writer to which the string is to be emitted
    
  post:
    s has been emitted to w

  test:
    s = None
    s = ""
      w = None
      w = valid Writer
    s = "fred"
  """
  type_checking2_0.check_derivative(s,unicoding3_0.String)
  type_checking2_0.check_derivative(w,unicode_io.Writer)
  i = 0
  while i < unicoding3_0.length_of(s):
    unicode_io.write(unicoding3_0.code_point_at(s,i),w)
    i += 1
  

def emit_start_tag(s,w):
  """ 
  pre:
    s = name of start tag, as lower-case Unicode code points
    w = Unicode stream to which start tag is to be emitted
    
  post:
    start tag element has been written to w

  syntax:
    StartTag(Name)  ::=  < Name OptionalSpaces >

  test:
    s = None
    s = valid non-empty String
      w = None
      w = valid Writer
  """
  type_checking2_0.check_derivative(s,unicoding3_0.String)
  type_checking2_0.check_derivative(w,unicode_io.Writer)
  unicode_io.write(ord('<'),w)
  i = 0
  while i < unicoding3_0.length_of(s):
    unicode_io.write(unicoding3_0.code_point_at(s,i),w)
    i += 1
  unicode_io.write(ord('>'),w)



# Test program for Contractor for Unicode input and output.

# author R.N.Bosworth

# version 27 Feb 2023  16:03

from emily0_9 import unicode_io
from guibits1_0 import unicoding3_0
"""
Copyright (C) 2017,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

print("Tests of unicode_io.new_input_reader", end=' ')
#r = unicode_io.new_input_reader("motorbike?%&*")
#r = unicode_io.new_input_reader("motorbike")
r = unicode_io.new_input_reader("motorbike.test")
print("OK")

print("Tests of unicode_io.new_string_reader", end=' ')
#r = unicode_io.new_string_reader(None)
#r = unicode_io.new_string_reader("hello")
r = unicode_io.new_string_reader(unicoding3_0.string_of("hello"))
print("OK")

print("Tests of unicode_io.read", end=' ')
r = unicode_io.new_input_reader("motorbike.test")
c = unicode_io.read(r)
assert c == ord('x')
assert unicode_io.read(r) == ord('y')
assert unicode_io.read(r) == ord('z')
assert unicode_io.read(r) == unicode_io.END_OF_STREAM
r = unicode_io.new_string_reader(unicoding3_0.string_of("\u0000a\uffff\U00010000\U0010ffff"))
assert unicode_io.read(r) == 0
assert unicode_io.read(r) == ord('a')
assert unicode_io.read(r) == 0xffff
cp = unicode_io.read(r)
assert cp == 0x10000
assert unicode_io.read(r) == 0x10ffff
assert unicode_io.read(r) == unicode_io.END_OF_STREAM
print("OK")

print("Tests of unicode_io.new_output_writer", end=' ')
#w = unicode_io.new_output_writer(None)
#w = unicode_io.new_output_writer("motorbike?%&*")
w = unicode_io.new_output_writer("pushbike.test")
print("OK")

print("Tests of unicode_io.new_string_writer", end=' ')
w = unicode_io.new_string_writer()
print("OK")

print("Tests of unicode_io.get_string", end=' ')
#s = unicode_io.get_string(None)
w = unicode_io.new_string_writer()
unicode_io.write(ord('b'),w)
s = unicode_io.get_string(w)
assert unicoding3_0.length_of(s) == 1
assert unicoding3_0.code_point_at(s,0) == ord('b')
print("OK")

print("Tests of unicode_io.write", end=' ')
w = unicode_io.new_output_writer("pushbike.test")
unicode_io.write(ord('x'),w)
unicode_io.write(ord('y'),w)
unicode_io.write(ord('z'),w)
#unicode_io.write(0x110000,w);
unicode_io.write(unicode_io.END_OF_STREAM,w)
#s = unicode_io.get_string(w)
w = unicode_io.new_string_writer()
#unicode_io.write(-2,w);
unicode_io.write(0x0000,w)
unicode_io.write(unicode_io.END_OF_STREAM,w)
s = unicode_io.get_string(w)
assert unicoding3_0.length_of(s) == 1
assert unicoding3_0.code_point_at(s,0) == 0x0000
w = unicode_io.new_string_writer()
unicode_io.write(ord('a'),w)
s = unicode_io.get_string(w)
assert unicoding3_0.length_of(s) == 1
assert unicoding3_0.code_point_at(s,0) == ord('a')
w = unicode_io.new_string_writer()
unicode_io.write(0xffff,w)
s = unicode_io.get_string(w)
assert unicoding3_0.length_of(s) == 1
assert unicoding3_0.code_point_at(s,0) == 0xffff
w = unicode_io.new_string_writer()
unicode_io.write(0x10000,w)
s = unicode_io.get_string(w)
assert unicoding3_0.length_of(s) == 1
assert unicoding3_0.code_point_at(s,0) == 0x10000
w = unicode_io.new_string_writer()
unicode_io.write(0x10ffff,w)
s = unicode_io.get_string(w)
assert unicoding3_0.length_of(s) == 1
assert unicoding3_0.code_point_at(s,0) == 0x10ffff
#unicode_io.write(0x110000,w);
w = unicode_io.new_string_writer()
unicode_io.write(unicode_io.END_OF_STREAM,w)
s = unicode_io.get_string(w)
assert unicoding3_0.length_of(s) == 0
print("OK")

print("")  
print("All tests OK")



# Test program for Contractor which deals with persistent data for Emily.

import io
import json
import os
from emily0_9 import persisting
from guibits1_0 import type_checking2_0, window_bounding

# author R.N.Bosworth

# version 28 Feb 23  18:29
""" 
Copyright (C) 2016,2017,2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

"""
Run with EmilyPersistents.json not present in user directory,
then with EmilyPersistents.json present in user directory
"""
print("Tests of persisting.get_current_file_name")
p = persisting.get_persistents()
print("  persisting.get_current_file_name(p)="+ \
     str(persisting.get_current_file_name(p)))
print("  If file name non-None re-run test program with EmilyPersistents.json absent from user directory")
assert persisting.get_current_file_name(p) == None
fn = "abc\u00010000\u00010001\u00010002def"
persisting.set_current_file_name(p,fn)
fn2 = persisting.get_current_file_name(p)
assert fn == fn2
print("OK")

print("Tests of persisting.get_menu_font_size", end=' ')
p = persisting.get_persistents()
assert persisting.get_menu_font_size(p) == 16.0
print("OK")

print("Tests of persisting.get_splash_screen_on", end=' ')
p = persisting.get_persistents()
assert persisting.get_splash_screen_on(p) == False
print("OK")

print("Tests of persisting.get_text_font_name", end=' ')
p = persisting.get_persistents()
assert persisting.get_text_font_name(p) == "Times New Roman"
print("OK")

print("Tests of persisting.get_text_font_size", end=' ')
p = persisting.get_persistents()
assert persisting.get_text_font_size(p) == 12.0
print("OK")

print("Tests of persisting.get_window_bounds", end=' ')
p = persisting.get_persistents()
wb = persisting.get_window_bounds(p)
assert window_bounding.get_x(wb) == 72.0
assert window_bounding.get_y(wb) == 72.0
assert window_bounding.get_width(wb) == 576.0
assert window_bounding.get_height(wb) == 288.0
print("OK")

print("Tests of persisting.get_zoom_factor", end=' ')
p = persisting.get_persistents()
assert persisting.get_zoom_factor(p) == 1.0
print("OK")

print("Tests of persisting._save_persistents")
p = persisting.Persistents()
persisting._save_persistents()
input("check EmilyPersistents.json in user's default directory - should be original values")
persisting.set_current_file_name(p,"fred")
persisting.set_zoom_factor(p,100.0)
persisting._save_persistents()
input("check EmilyPersistents.json in user's default directory - should be new values")
print("OK")

print("Tests of persisting.set_current_file_name")
p = persisting.get_persistents()
persisting.set_current_file_name(p,None)
input("check EmilyPersistents.json in user's default directory - current file name should be null")
persisting.set_current_file_name(p,"C:\\")
assert persisting.get_current_file_name(p) == "C:\\"
input("check EmilyPersistents.json in user's default directory - current file name should be \"C:\\\\\"")
print("OK")

print("Tests of persisting.set_menu_font_size")
p = persisting.get_persistents()
persisting.set_menu_font_size(p,18.0)
assert persisting.get_menu_font_size(p) == 18.0
input("check EmilyPersistents.json in user's default directory - menu font size should be 18.0")
print("OK")

print("Tests of persisting.set_splash_screen_on")
p = persisting.get_persistents()
persisting.set_splash_screen_on(p,True)
assert persisting.get_splash_screen_on(p) == True
input("check EmilyPersistents.json in user's default directory - splash screen should be on")
print("OK")

print("Tests of set_text_font_name")
p = persisting.get_persistents()
persisting.set_text_font_name(p,"John Major")
assert persisting.get_text_font_name(p) == "John Major"
input("check EmilyPersistents.json in user's default directory - text font name should be \"John Major\"")
print("OK")

print("Tests of persisting.set_text_font_size")
p = persisting.get_persistents()
persisting.set_text_font_size(p,99.0)
assert persisting.get_text_font_size(p) == 99.0
input("check EmilyPersistents.json in user's default directory - font size should be 99.0")
print("OK")

print("Tests of set_window_bounds")
p = persisting.get_persistents()
wb = window_bounding.new_bounds()
window_bounding.set_x(wb,5.0)
window_bounding.set_y(wb,6.0)
window_bounding.set_width(wb,7.0)
window_bounding.set_height(wb,8.0)
persisting.set_window_bounds(p,wb)
wb2 = persisting.get_window_bounds(p)
assert window_bounding.get_x(wb2) == 5.0
assert window_bounding.get_y(wb2) == 6.0
assert window_bounding.get_width(wb2) == 7.0
assert window_bounding.get_height(wb2) == 8.0
input("check EmilyPersistents.json in user's default directory - window bounds should be 5,6,7,8")
print("OK")

print("Tests of set_zoom_factor")
p = persisting.get_persistents()
persisting.set_zoom_factor(p,2.0)
assert persisting.get_zoom_factor(p) == 2.0
input("check EmilyPersistents.json in user's default directory - zoom factor should be 2.0")
print("OK")

print("Tests of persisting.get_persistents", end=' ')
p = persisting.get_persistents()
assert p != None
g2 = persisting.get_persistents()
assert g2 != None
assert p == g2
print("OK")

print("Tests of paper dimensions", end=' ')
p = persisting.get_persistents()
assert persisting.get_paper_height(p) == 841.0
assert persisting.get_paper_width(p) == 592.0
assert persisting.get_horizontal_indent(p) == 72.0
assert persisting.get_vertical_indent(p) == 72.0
print("OK")

print("")
print("All tests OK")



# Test program for Contractor to print texts.

# author R.N.Bosworth

# version 16 Mar 23  15:06

from emily0_9 import rendering, text_printing, texting
from guibits1_0 import coloring, font_styling, printing
from guibits1_0 import unicoding3_0, windowing, writing

""" 
Copyright (C) 2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------
   
print("Tests of text_printing._de_escape", end=' ')
s1 = unicoding3_0.new_string()
s2 = unicoding3_0.new_string()
text_printing._de_escape(s1,s2)
assert unicoding3_0.equals(s1,unicoding3_0.new_string())
s1 = unicoding3_0.string_of("ab")
s2 = unicoding3_0.string_of("1&lt;2")
text_printing._de_escape(s1,s2)
assert unicoding3_0.equals(s1,unicoding3_0.string_of("ab1<2"))
print("OK")
  
# tests requiring a window
# ------------------------
  
def window_closing(win):
  print("start of window_closing")
  print("Tests of text_printing._try_line", end=' ')
  fss = font_styling.new_font_styles()
  pg = rendering.new_page(win,216.0,168.0,72.0,72.0,"Courier New",fss,12.0,coloring.BLACK)
  #text_printing._try_line(pg)
  pj = printing.set_page_dimensions(216.0,168.0,72.0)
  pg._current_line = unicoding3_0.string_of("Test of text_printing._try_line")
  text_printing._try_line(pj,pg)
  pg._page_position.y_offset += pg._font_size
  pg._current_line = unicoding3_0.string_of("Test of text_printing._try_line just on page")
  text_printing._try_line(pj,pg)
  pg._page_position.y_offset += pg._font_size
  pg._current_line = unicoding3_0.string_of("Test of text_printing._try_line line off page")
  text_printing._try_line(pj,pg)
  printing.end_printing(pj)
  print("OK")
  
  print("Tests of text_printing._print_line", end=' ')
  fss = font_styling.new_font_styles()
  pg = rendering.new_page(win,216.0,180.0,72.0,72.0,"Courier New",fss,12.0,coloring.BLACK)
  pj = printing.set_page_dimensions(216.0,180.0,72.0)
  pg._current_line = unicoding3_0.string_of("fred1")
  pg._line_width = writing.width_in_points_of(win,unicoding3_0.python_string_of(pg._current_line),"Courier New",fss,12.0)
  pg._current_alignment = texting.Alignment.BEGIN
  text_printing._print_line(pj,pg)
  pg._current_line = unicoding3_0.string_of("fred2")
  pg._line_width = writing.width_in_points_of(win,unicoding3_0.python_string_of(pg._current_line),"Courier New",fss,12.0)
  pg._current_alignment = texting.Alignment.END
  text_printing._print_line(pj,pg)
  pg._current_line = unicoding3_0.string_of("fred3")
  pg._line_width = writing.width_in_points_of(win,unicoding3_0.python_string_of(pg._current_line),"Courier New",fss,12.0)
  pg._current_alignment = texting.Alignment.MIDDLE
  text_printing._print_line(pj,pg)
  printing.end_printing(pj)
  print("OK")
  
  print("Tests of text_printing._print_token", end=' ')
  fss = font_styling.new_font_styles()
  pg = rendering.new_page(win,216.0,288.0,36.0,36.0,"Courier New",fss,12.0,coloring.BLACK)
  
  # initialize printing
  pj = printing.set_page_dimensions(216.0,288.0,36.0)
  # start of print
  pg._current_alignment = texting.Alignment.BEGIN
  pg._page_position.x_offset = 0.0
  pg._page_position.y_offset = 0.0
  pg._current_line = unicoding3_0.new_string()
  pg._line_width = 0.0
  text_printing._print_token(pj,unicoding3_0.string_of("The"),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("cat"),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.BEGIN
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.BEGIN
  text_printing._print_token(pj,unicoding3_0.string_of("sat"),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.BEGIN
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.BEGIN
  text_printing._print_token(pj,unicoding3_0.string_of("on"),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.BEGIN
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.BEGIN
  text_printing._print_token(pj,unicoding3_0.string_of("the"),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.BEGIN
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.BEGIN
  text_printing._print_token(pj,unicoding3_0.string_of("mat."),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.BEGIN
  text_printing._print_token(pj,unicoding3_0.string_of("\n"),texting.Alignment.MIDDLE,pg)
  assert pg._current_alignment == texting.Alignment.MIDDLE
  text_printing._print_token(pj,unicoding3_0.string_of("\n"),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.END
  printing.end_printing(pj)
  
  # start of print
  pj = printing.set_page_dimensions(216.0,288.0,36.0)
  pg._current_alignment = texting.Alignment.BEGIN
  pg._page_position.x_offset = 0.0
  pg._page_position.y_offset = 0.0
  pg._current_line = unicoding3_0.new_string()
  pg._line_width = 0.0
  text_printing._print_token(pj,unicoding3_0.string_of("The"),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("cat"),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("sat"),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("on"),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("the"),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.END,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("mat."),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.BEGIN
  text_printing._print_token(pj,unicoding3_0.string_of("\n"),texting.Alignment.MIDDLE,pg)
  assert pg._current_alignment == texting.Alignment.MIDDLE
  text_printing._print_token(pj,unicoding3_0.string_of("\n"),texting.Alignment.END,pg)
  assert pg._current_alignment == texting.Alignment.END
  printing.end_printing(pj)
  
  # start of print
  pj = printing.set_page_dimensions(216.0,288.0,36.0)
  pg._current_alignment = texting.Alignment.BEGIN
  pg._page_position.x_offset = 0.0
  pg._page_position.y_offset = 0.0
  pg._current_line = unicoding3_0.new_string()
  pg._line_width = 0.0
  text_printing._print_token(pj,unicoding3_0.string_of("The"),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("cat"),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("sat"),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("on"),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("the"),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of(" "),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("\n"),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("\n"),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("mat."),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("<end>"),texting.Alignment.BEGIN,pg)
  assert pg._current_alignment == texting.Alignment.BEGIN
  printing.end_printing(pj)
  
  # start of print
  pj = printing.set_page_dimensions(216.0,288.0,36.0)
  pg._current_alignment = texting.Alignment.BEGIN
  pg._page_position.x_offset = 0.0
  pg._page_position.y_offset = 0.0
  pg._current_line = unicoding3_0.new_string()
  pg._line_width = 0.0
  text_printing._print_token(pj,unicoding3_0.string_of("the"),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("<end>"),texting.Alignment.BEGIN,pg)
  # release printer
  printing.end_printing(pj)
  """
  # start of print  (bad token)
  pj = printing.set_page_dimensions(216.0,288.0,36.0)
  pg._current_alignment = texting.Alignment.BEGIN
  pg._page_position.x_offset = 0.0
  pg._page_position.y_offset = 0.0
  pg._current_line = unicoding3_0.new_string()
  pg._line_width = 0.0
  text_printing._print_token(pj,unicoding3_0.string_of("the"),texting.Alignment.BEGIN,pg)
  text_printing._print_token(pj,unicoding3_0.string_of("<exit>"),texting.Alignment.BEGIN,pg)
  # release printer
  printing.end_printing(pj)
  """
  print("OK")
  
  print("Tests of text_printing.print_text",end=' ')
  fss = font_styling.new_font_styles()
  t = texting.new_text()
  text_printing.print_text(win,t,"Courier New",12.0,fss)
  t = texting.new_text()
  texting.set_alignment(t,texting.Alignment.END)
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('1'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('2'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('3'))
  texting.insert_code_point(t,ord('.'))
  text_printing.print_text(win,t,"Courier New",12.0,fss)

  t = texting.new_text()
  texting.set_alignment(t,texting.Alignment.MIDDLE)
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('1'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('2'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('3'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('4'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('5'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('6'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('7'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('8'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('9'))
  texting.insert_code_point(t,ord('.'))
  text_printing.print_text(win,t,"Courier New",12.0,fss)
  
  t = texting.new_text()
  texting.set_alignment(t,texting.Alignment.BEGIN)
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('1'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('2'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('3'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('4'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('5'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('6'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('7'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('8'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('9'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('1'))
  texting.insert_code_point(t,ord('0'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('1'))
  texting.insert_code_point(t,ord('1'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('1'))
  texting.insert_code_point(t,ord('2'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('1'))
  texting.insert_code_point(t,ord('3'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('1'))
  texting.insert_code_point(t,ord('4'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('1'))
  texting.insert_code_point(t,ord('5'))
  texting.insert_code_point(t,ord(' '))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('h'))
  texting.insert_code_point(t,ord('u'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('a'))
  texting.insert_code_point(t,ord('r'))
  texting.insert_code_point(t,ord('b'))
  texting.insert_code_point(t,ord('1'))
  texting.insert_code_point(t,ord('6'))
  texting.insert_code_point(t,ord('.'))
  text_printing.print_text(win,t,"Courier New",12.0,fss)
  print("OK")
  print("")
  print("All tests OK")
  return True

print("about to open new window")
win = windowing.new_window(12.0,"Window for test of text_printing",800.0,500.0,1.0)
windowing.show(win,None,window_closing)
  



# Test program for Contractor for the basic grammar of HTML and CSS.

# author R.N.Bosworth

# version 8 Mar 23  15:40

from emily0_9 import html_css_parsing, looking_ahead, unicode_io
from guibits1_0 import unicoding3_0

""" 
Copyright (C) 2017,2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

print("Tests of accept", end=' ')
#html_css_parsing.accept('A',None)
#html_css_parsing.accept(ord('A'),None)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("B"))
#html_css_parsing.accept(ord('A'),la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("A"))
html_css_parsing.accept(ord('A'),la)
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
print("OK")

print("Tests of accept_optional_spaces", end=' ')
#html_css_parsing.accept_optional_spaces(None)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("hello"))
html_css_parsing.accept_optional_spaces(la)
assert looking_ahead.current_symbol_of(la) == ord('h')
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("  hello"))
html_css_parsing.accept_optional_spaces(la)
assert looking_ahead.current_symbol_of(la) == ord('h')
print("OK")

print("Tests of accept_separator", end=' ')
#html_css_parsing.accept_separator(None);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of(""))
#html_css_parsing.accept_separator(la);
assert looking_ahead.current_symbol_of(la) == unicode_io.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("a"))
#html_css_parsing.accept_separator(la);
assert looking_ahead.current_symbol_of(la) == ord('a')
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of(" "))
html_css_parsing.accept_separator(la)
assert looking_ahead.current_symbol_of(la) == unicode_io.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("   "))
html_css_parsing.accept_separator(la)
assert looking_ahead.current_symbol_of(la) == unicode_io.END_OF_STREAM
print("OK")

print("Tests of parse_name", end=' ')
#html_css_parsing.parse_name(None)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("%"))
#s = html_css_parsing.parse_name(la)
la = looking_ahead.lookahead_of_string(unicoding3_0.new_string())
#s = html_css_parsing.parse_name(la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("f"))
s = html_css_parsing.parse_name(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("f"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("fred"))
s = html_css_parsing.parse_name(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("fred"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("FRED"))
#s = html_css_parsing.parse_name(la)
print("OK")

print("Tests of accept_name", end=' ')
#html_css_parsing.accept_name(None,None)
#html_css_parsing.accept_name(unicoding3_0.string_of("fred"),None)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("jim"))
#html_css_parsing.accept_name(unicoding3_0.string_of("fred"),la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("fred"))
html_css_parsing.accept_name(unicoding3_0.string_of("fred"),la)
print("OK")

print("")
print("All tests OK")



# Contractor for the basic grammar of HTML and CSS.

# author R.N.Bosworth

# version 8 Mar 23  15:42

from guibits1_0 import type_checking2_0, unicoding3_0

from .import code_point_testing, looking_ahead, unicode_io

""" 
Copyright (C) 2017,2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""
""" 
  HTML CSS Grammar version 17.3.18   15:39

  This grammar defines the basic HTML constructs common to HTML5 and CSS.

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::=  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive, unlike HTML5.
  
  
  OptionalSpaces  ::=  SpaceChar*

  SpaceChar  ::=  (SPACE | TAB | LF | FF | CR)
                  (U+0020 | U+0009 | U+000A | U+000C | U+000D )
                  
  Separator  ::==  SpaceChar OptionalSpaces
  
  Name  ::=  NameChar NameChar*

  NameChar  ::=  ( a .. z | 0 .. 9 )
                 ( U+0061 .. U+007A | U+0030 .. U+0039 )
                 
"""

# exposed procedures
# ------------------

def accept(cp,la):
  """ 
  pre:
    cp = unicode code point which is to be accepted, as an int
    la = looking_ahead.Lookahead which is being parsed
    
  post:
    either:
      cp has been accepted from la
      la has been advanced past cp
    or:
      RuntimeException has been thrown

  test:
    cp = None
    cp = 'A'
      la = None
      la = "B"
      la = "A"
  """
  type_checking2_0.check_identical(cp,int)
  type_checking2_0.check_derivative(la,looking_ahead.Lookahead)
  if looking_ahead.current_symbol_of(la) != cp:
    raise Exception("unexpected code point, expected: U+" + format(cp,'04x') +" found: U+" + format(looking_ahead.current_symbol_of(la),'04x'))
  else:
    looking_ahead.advance(la)
    
  
def accept_name(name,la):
  """ 
  pre:
    name = name to be accepted, as unicoding3_0.String
    la = looking_ahead.Lookahead to be parsed
    
  post:
    either:
      name has been accepted,
    or:
      a parse exception as been raised

  syntax:
    Name  ::=  NameChar NameChar*

  test:
    name = None
      la = None
    name = "jim" (not expected)
      la = None
           valid Lookahead
    name = "fred" (expected)
      la = valid Lookahead
  """
  type_checking2_0.check_derivative(name,unicoding3_0.String)
  type_checking2_0.check_derivative(la,looking_ahead.Lookahead)
  s = parse_name(la)
  if  not unicoding3_0.equals(s,name):
    raise Exception("Unexpected name: expected \"" + unicoding3_0.python_string_of(name) + "\", found \"" + unicoding3_0.python_string_of(s) + "\"")
    

def accept_optional_spaces(la):
  """ 
  pre:
    la = looking_ahead.Lookahead to be parsed
    
  post:
    zero or more space characters have been consumed from la

  syntax:
    OptionalSpaces  ::=  SpaceChar*

  test:
    la = None
    la = "hello"
    la = "  hello"
  """
  type_checking2_0.check_derivative(la,looking_ahead.Lookahead)
  cp = looking_ahead.current_symbol_of(la)
  while code_point_testing.is_space_char(cp):
    looking_ahead.advance(la)
    cp = looking_ahead.current_symbol_of(la)
    

def accept_separator(la):
  """ 
  pre:
    la = looking_ahead.Lookahead which is being parsed
    
  post:
    either:
      accepts separator of one or more Space code points from la
    or:
      an Exception has been raised

  syntax:
    Separator  ::==  SpaceChar OptionalSpaces

  test:
    la = None
         ""
         "a"
         " "
         "   "
  """
  type_checking2_0.check_derivative(la,looking_ahead.Lookahead)
  if  not code_point_testing.is_space_char(looking_ahead.current_symbol_of(la)):
    raise Exception("unexpected code point in HTML separator: U+" +format(looking_ahead.current_symbol_of(la),'04x'))
  accept_optional_spaces(la)
  

def parse_name(la):
  """ 
  pre:
    la = looking_ahead.Lookahead which is being parsed
    
  post:
    either:
      returns HTML name as lower-case unicoding3_0.String
      lookahead has been advanced past Name
    or:
      an Exception has been raised

  syntax:
    Name  ::=  NameChar NameChar*

  test:
    la = None
         "%"
         ""
         "f"
         "fred"
         "FRED"
  """
  type_checking2_0.check_derivative(la,looking_ahead.Lookahead)
  s = unicoding3_0.new_string()
  if  not code_point_testing.is_name_char(looking_ahead.current_symbol_of(la)):
    raise Exception("unexpected code point in HTML name: U+" + format(looking_ahead.current_symbol_of(la),'04x'))
  unicoding3_0.append(s,looking_ahead.current_symbol_of(la))
  looking_ahead.advance(la)
  while code_point_testing.is_name_char(looking_ahead.current_symbol_of(la)):
    unicoding3_0.append(s,looking_ahead.current_symbol_of(la))
    looking_ahead.advance(la)
  return s



# Test program for Contractor which converts text to tokens.

# author R.N.Bosworth

# version 8 Mar 23  10:53

from emily0_9 import text_tokenizing, texting
from guibits1_0 import unicoding3_0

""" 
Copyright (C) 2018,2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

def _equals(s1,s2):
  return unicoding3_0.equals(s1,unicoding3_0.string_of(s2))


print("Tests of _parse_token", end=' ')
t = texting.new_text()
texting.insert_code_point(t,ord(' '))
texting.insert_code_point(t,ord('f'))
texting.insert_code_point(t,ord('r'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('d'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('x'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('r'))
texting.insert_code_point(t,ord('t'))
texting.set_cursor_start(t)
s = text_tokenizing._parse_token(t)
assert _equals(s," ")
s = text_tokenizing._parse_token(t)
assert _equals(s,"fred")
s = text_tokenizing._parse_token(t)
assert _equals(s,"<b>")
s = text_tokenizing._parse_token(t)
assert _equals(s,"x")
s = text_tokenizing._parse_token(t)
assert _equals(s,"\n")
s = text_tokenizing._parse_token(t)
assert _equals(s,"bert")
#s = text_tokenizing._parse_token(t) # end of text - should fail
print("OK")

print("Tests of get_next_token", end=' ')
#s = text_tokenizing.get_next_token(None)
t = texting.new_text()
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('i'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord(' '))
texting.insert_code_point(t,ord('f'))
texting.insert_code_point(t,ord('r'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('d'))
texting.set_cursor_start(t)
s = text_tokenizing.get_next_token(t)
assert _equals(s,"<i>")
s = text_tokenizing.get_next_token(t)
assert _equals(s," ")
s = text_tokenizing.get_next_token(t)
assert _equals(s,"fred")
s = text_tokenizing.get_next_token(t)
assert _equals(s,"<end>")
s = text_tokenizing.get_next_token(t)
assert _equals(s,"<end>")
print("OK")

print("")
print("All tests OK")



# Test of Contractor for emitting html elements.

# author R.N.Bosworth

# version 2 Mar 2023  14:45

""" 
Contractor for emitting html elements.

Copyright (C) 2017,2018,2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

from emily0_9 import html_emitting, unicode_io
from guibits1_0 import unicoding3_0

# test program
# ------------

print("Tests of html_emitting.emit_end_tag", end=' ')
#html_emitting.emit_end_tag(None,None)
#html_emitting.emit_end_tag(unicoding3_0.string_of("html"),None)
w = unicode_io.new_string_writer()
html_emitting.emit_end_tag(unicoding3_0.string_of("html"),w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("</html>"))
print("OK")

print("Tests of html_emitting.emit_indent", end=' ')
#html_emitting.emit_indent(None)
w = unicode_io.new_string_writer()
html_emitting.emit_indent(w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("  "))
print("OK")

print("Tests of html_emitting.emit_new_line", end=' ')
#html_emitting.emit_new_line(None)
w = unicode_io.new_string_writer()
html_emitting.emit_new_line(w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("\n"))
print("OK")

print("Tests of html_emitting.emit_start_tag", end=' ')
#html_emitting.emit_start_tag(None,None)
#html_emitting.emit_start_tag(unicoding3_0.string_of("html"),None)
w = unicode_io.new_string_writer()
html_emitting.emit_start_tag(unicoding3_0.string_of("html"),w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("<html>"))
print("OK")

print("Tests of html_emitting.emit_string", end=' ')
#html_emitting.emit_string(None,None)
s = unicoding3_0.new_string()
#html_emitting.emit_string(s,None)
w = unicode_io.new_string_writer()
html_emitting.emit_string(s,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of(""))
s = unicoding3_0.string_of("fred")
w = unicode_io.new_string_writer()
html_emitting.emit_string(s,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("fred"))
print("OK")

print("")
print("All tests OK")



# Test of Contractor which allows the client to advance and retreat through HTML # text, as it appears on the screen,
# ignoring tags and dealing with escaped code points.

# author R.N.Bosworth

# version 2 Mar 23  15:12

"""
Copyright (C) 2018,2021,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

from emily0_9 import html_texting, texting

# test program
# ------------

print("Tests of advance", end=' ')
# assert html_texting.advance(None)
t = texting.new_text()
texting.set_cursor_start(t)
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('i'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('1'))
texting.insert_code_point(t,ord('&'))
texting.insert_code_point(t,ord('l'))
texting.insert_code_point(t,ord('t'))
texting.insert_code_point(t,ord(';'))
texting.insert_code_point(t,ord('2'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('/'))
texting.insert_code_point(t,ord('i'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('/'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('x'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('/'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('>'))
texting.set_cursor_start(t)
assert html_texting.advance(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 7
assert html_texting.advance(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 11
assert html_texting.advance(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 12
assert html_texting.advance(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 21
assert not html_texting.advance(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 28
print("OK")

print("Tests of delete_after", end=' ')
#assert html_texting.delete_after(None)
t = texting.new_text()
texting.set_cursor_start(t)
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('i'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('1'))
texting.insert_code_point(t,ord('&'))
texting.insert_code_point(t,ord('l'))
texting.insert_code_point(t,ord('t'))
texting.insert_code_point(t,ord(';'))
texting.insert_code_point(t,ord('2'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('/'))
texting.insert_code_point(t,ord('i'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('/'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('x'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('/'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('>'))
texting.set_cursor_start(t)
assert html_texting.delete_after(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 6
assert texting.current_code_point(t) == ord('&')
assert html_texting.delete_after(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 6
assert texting.current_code_point(t) == ord('2')
assert html_texting.delete_after(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 6
assert texting.current_code_point(t) == ord('<')
assert html_texting.delete_after(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 14
assert texting.current_code_point(t) == ord('<')
assert not html_texting.delete_after(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 21
assert texting.current_code_point(t) == texting.END_OF_TEXT
print("OK")

print("Tests of retreat", end=' ')
#assert html_texting.retreat(None)
t = texting.new_text()
texting.set_cursor_start(t)
texting.insert_code_point(t,ord('x'))
texting.insert_code_point(t,ord(';'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('i'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('1'))
texting.insert_code_point(t,ord('&'))
texting.insert_code_point(t,ord('l'))
texting.insert_code_point(t,ord('t'))
texting.insert_code_point(t,ord(';'))
texting.insert_code_point(t,ord(';'))
texting.insert_code_point(t,ord(' '))
texting.insert_code_point(t,ord(';'))
texting.insert_code_point(t,ord('2'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('/'))
texting.insert_code_point(t,ord('i'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('/'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('y'))
texting.set_cursor_end(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 26
assert html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 25
assert html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 16
assert html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 15
assert html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 14
assert html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 13
assert html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 9
assert html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 8
assert html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 1
assert html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 0
assert not html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 0
t = texting.new_text()
texting.set_cursor_start(t)
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('l'))
texting.insert_code_point(t,ord('i'))
texting.insert_code_point(t,ord('g'))
texting.insert_code_point(t,ord('n'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('x'))
texting.set_cursor_end(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 8
assert html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 7
assert not html_texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 0
print("OK")

print("")  
print("All tests OK")



from . import css_emitting
from . import html_emitting
from . import texting
from . import text_emitting
from guibits1_0 import type_checking2_0
from . import unicode_io
from guibits1_0 import unicoding3_0

# author R.N.Bosworth

# version 9 Mar 23  15:27
""" 
Contractor for emitting an MLE document.

Copyright (C) 2017,2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""
""" 
  MLE Grammar version 17.3.18  15:53
  
  This grammar is a reduced version of the HTML grammar given in WHATWG's HTML5 website.

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::=  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive, unlike HTML5.
  
  Parameterization is used to distinguish specialized non-terminals such as tags.

  
  MleDocument  ::=  OptionalSpaces Doctype OptionalSpaces HtmlElement OptionalSpaces

  Doctype  ::=  < ! d o c t y p e Separator h t m l OptionalSpaces >
  
  HtmlElement  ::=  StartTag(html) OptionalSpaces HeadElement OptionalSpaces BodyElement 
                      OptionalSpaces EndTag(html)

  HeadElement  ::=  StartTag(head) OptionalSpaces MetaTag OptionalSpaces TitleElement 
                      OptionalSpaces StyleElement OptionalSpaces EndTag(head)
                     
  MetaTag  ::=  < m e t a Separator c h a r s e t OptionalSpaces = OptionalSpaces " U T F - 8 " 
                  OptionalSpaces >

  TitleElement  ::=  StartTag(title) Text EndTag(title)

  Text  ::=  EscapedUnicodeCodePoint*

  StyleElement  ::=  StartTag(style) OptionalSpaces StyleSheet OptionalSpaces EndTag(style)
                      (for StyleSheet syntax see CSSGrammar.txt)

  BodyElement  ::=  
    StartTag(body) OptionalSpaces (ParagraphOrBreak  OptionalSpaces)* 
	    EndTag(body)
    (note: paragraphs are separated by a blank line in the Emily text)
    
  ParagraphOrBreak  ::=  (Paragraph | BreakTag)
                    
  BreakTag  ::=  StartTag(br)

  Paragraph  ::=  ParagraphStartTag OptionalSpaces ParagraphText EndTag(p)
  
  ParagraphStartTag  ::=  < p ( > | Separator ( > | ClassAttribute OptionalSpaces > )) (ci)
  
  ClassAttribute  ::==  c l a s s OptionalSpaces = 
                          OptionalSpaces " ClassName "
  
  ClassName  ::==  ( b e g i n |  m i d d l e | e n d | j u s t i f y )
                   
  ParagraphText  ::= (EscapedUnicodeCodePoint | BreakTag OptionalSpaces)*
                      (note: in the Emily internal text,
                       code points remain escaped,
                       words are separated by a single space, 
                       and each BreakTag is followed by a newline.)
                       
  Note: Rules for OptionalSpaces and Separator are given in the HTMLCSSGrammar.  Rules for EscapedUnicodeCodePoint, StartTag and EndTag are given in the HTMLGrammar.
   
  Simplifications from HTML5:
  
   1. No legacy string in Doctype.
   
   2. Html start and end tags mandatory.
   
   3. No attributes in start tag (for the moment).
   
   4. No solidus in start tag.
   
   5. Head contains only title.
   
   6. Body contains only paragraphs and break tags.
   
   7. Paragraph has mandatory end tag.
   
   8. Comments not allowed (for security).
   
   9. Tags and attribute names are case-sensitive.
"""


# exposed procedures
# ------------------

def emit_emily_document(n,d,v,t,w):
  """ 
  pre:
    n = font name for this document's body, as unicoding3_0.String
    d = font size for this document's body in range 0.0 - 99.0, as float
    v = version number of this version of Emily, as unicoding3_0.String
    t = texting.Text which is to be emitted
    w = unicode_io.Writer on which t is to be emitted

  test:
    n = None
    n = valid font name
      d = None
      d = -0.1
      d = 99.1
      d =10.0
        v = None
        v = valid version number string
        t = None
        t = empty text
          w = None
          w = valid Writer
        t = non-empty text
          w = valid Writer
  """
  type_checking2_0.check_derivative(n,unicoding3_0.String)
  type_checking2_0.check_identical(d,float)
  if d < 0.0 or d > 99.0:
    raise Exception("Attempt to emit text with font size < 0.0 or >99.0: d="+str(d))
  type_checking2_0.check_derivative(v,unicoding3_0.String)
  type_checking2_0.check_derivative(t,texting.Text)
  type_checking2_0.check_derivative(w,unicode_io.Writer)
  html_emitting.emit_start_tag(unicoding3_0.string_of("!doctype html"),w)
  html_emitting.emit_new_line(w)
  html_emitting.emit_start_tag(unicoding3_0.string_of("html"),w)
  html_emitting.emit_new_line(w)
  html_emitting.emit_start_tag(unicoding3_0.string_of("head"),w)
  html_emitting.emit_new_line(w)
  html_emitting. emit_indent(w)
  html_emitting.emit_start_tag(unicoding3_0.string_of("meta charset=\"UTF-8\""),w)
  html_emitting.emit_new_line(w)
  html_emitting. emit_indent(w)
  html_emitting.emit_start_tag(unicoding3_0.string_of("title"),w)
  vs = unicoding3_0.string_of("Emily version ")
  unicoding3_0.append_a_copy(vs,v)
  unicoding3_0.append_a_copy(vs,unicoding3_0.string_of(" Document"))
  html_emitting.emit_string(vs,w)
  html_emitting.emit_end_tag(unicoding3_0.string_of("title"),w)
  html_emitting.emit_new_line(w)
  html_emitting.emit_start_tag(unicoding3_0.string_of("style"),w)
  html_emitting.emit_new_line(w)
  css_emitting.emit_style_sheet(n,d,w)
  html_emitting.emit_end_tag(unicoding3_0.string_of("style"),w)
  html_emitting.emit_new_line(w)
  html_emitting.emit_end_tag(unicoding3_0.string_of("head"),w)
  html_emitting.emit_new_line(w)
  html_emitting.emit_start_tag(unicoding3_0.string_of("body"),w)
  html_emitting.emit_new_line(w)
  with t._my_lock:
    text_emitting.parse_text(t,w)
  html_emitting.emit_end_tag(unicoding3_0.string_of("body"),w)
  html_emitting.emit_new_line(w)
  html_emitting.emit_end_tag(unicoding3_0.string_of("html"),w)
  html_emitting.emit_new_line(w)
  unicode_io.write(unicode_io.END_OF_STREAM,w)
